"""
The core part of the SOTA model of CPSC2019,
branched, and has different scope (in terms of dilation) in each branch.
"""

import textwrap
from collections import OrderedDict
from copy import deepcopy
from itertools import repeat
from typing import Optional, Sequence, Union, List

import numpy as np
import torch
from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import Conv_Bn_Activation, DownSample
from ...utils.misc import list_sum, add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "MultiScopicCNN",
    "MultiScopicBasicBlock",
    "MultiScopicBranch",
]


if not hasattr(nn, "Dropout1d"):
    nn.Dropout1d = nn.Dropout  # added in pytorch 1.12


class MultiScopicBasicBlock(nn.Sequential, SizeMixin):
    """Basic building block of the CNN part of the SOTA model
    from CPSC2019 challenge (entry 0416).

    (conv -> activation) * N --> bn --> down_sample

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    scopes : Sequence[int]
        Scopes of the convolutional layers, via `dilation`.
    num_filters : int or Sequence[int]
        Number of filters of the convolutional layer(s).
    filter_lengths : int or Sequence[int]
        Filter length(s) (kernel size(s)) of the convolutional layer(s).
    subsample_length : int
        Subsample length (ratio) at the last layer of the block.

    """

    __name__ = "MultiScopicBasicBlock"

    def __init__(
        self,
        in_channels: int,
        scopes: Sequence[int],
        num_filters: Union[int, Sequence[int]],
        filter_lengths: Union[int, Sequence[int]],
        subsample_length: int,
        groups: int = 1,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__scopes = scopes
        self.__num_convs = len(self.__scopes)
        if isinstance(num_filters, int):
            self.__out_channels = list(repeat(num_filters, self.__num_convs))
        else:
            self.__out_channels = num_filters
            assert len(self.__out_channels) == self.__num_convs, (
                f"`scopes` indicates {self.__num_convs} convolutional layers, "
                f"while `num_filters` indicates {len(self.__out_channels)}"
            )
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, self.__num_convs))
        else:
            self.__filter_lengths = filter_lengths
            assert len(self.__filter_lengths) == self.__num_convs, (
                f"`scopes` indicates {self.__num_convs} convolutional layers, "
                f"while `filter_lengths` indicates {len(self.__filter_lengths)}"
            )
        self.__subsample_length = subsample_length
        self.__groups = groups
        self.config = CFG(deepcopy(config))

        conv_in_channels = self.__in_channels
        for idx in range(self.__num_convs):
            self.add_module(
                f"ca_{idx}",
                Conv_Bn_Activation(
                    in_channels=conv_in_channels,
                    out_channels=self.__out_channels[idx],
                    kernel_size=self.__filter_lengths[idx],
                    stride=1,
                    dilation=self.__scopes[idx],
                    groups=self.__groups,
                    norm=self.config.get("norm", self.config.get("batch_norm")),
                    # kw_bn=self.config.kw_bn,
                    activation=self.config.activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    bias=self.config.bias,
                ),
            )
            conv_in_channels = self.__out_channels[idx]
        self.add_module("bn", nn.BatchNorm1d(self.__out_channels[-1]))
        self.add_module(
            "down",
            DownSample(
                down_scale=self.__subsample_length,
                in_channels=self.__out_channels[-1],
                groups=self.__groups,
                # padding=
                norm=False,
                mode=self.config.subsample_mode,
            ),
        )
        if isinstance(self.config.dropout, dict):
            if self.config.dropout.type == "1d" and self.config.dropout.p > 0:
                self.add_module(
                    "dropout",
                    nn.Dropout1d(
                        p=self.config.dropout.p,
                        inplace=self.config.dropout.get("inplace", False),
                    ),
                )
            elif self.config.dropout.type is None and self.config.dropout.p > 0:
                self.add_module(
                    "dropout",
                    nn.Dropout(
                        p=self.config.dropout.p,
                        inplace=self.config.dropout.get("inplace", False),
                    ),
                )
        elif self.config.dropout > 0:
            self.add_module("dropout", nn.Dropout(self.config.dropout, inplace=False))

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        output = super().forward(input)
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            The batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the block.

        """
        _seq_len = seq_len
        for _, module in enumerate(self):
            if isinstance(module, (nn.Dropout, nn.Dropout1d, nn.BatchNorm1d)):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape

    def _assign_weights_lead_wise(
        self, other: "MultiScopicBasicBlock", indices: Sequence[int]
    ) -> None:
        """Assign weights lead-wise.

        This method is used to assign weights from a model with
        a superset of the current model's leads to the current model.

        Parameters
        ----------
        other : MultiScopicBasicBlock
            The model with a superset of the current model's leads.
        indices : Sequence[int]
            The indices of the leads of the current model in the
            superset model.

        Returns
        -------
        None

        """
        assert not any([isinstance(m, nn.LayerNorm) for m in self]) and not any(
            [isinstance(m, nn.LayerNorm) for m in other]
        ), "Lead-wise assignment of weights is not supported for the existence of `LayerNorm` layers"
        for blk, o_blk in zip(self, other):
            if isinstance(blk, Conv_Bn_Activation):
                blk._assign_weights_lead_wise(o_blk, indices)
            elif isinstance(blk, (nn.BatchNorm1d, nn.GroupNorm, nn.InstanceNorm1d)):
                if hasattr(blk, "num_features"):  # batch norm and instance norm
                    out_channels = blk.num_features
                else:  # group norm
                    out_channels = blk.num_channels
                units = out_channels // self.groups
                out_indices = list_sum(
                    [[i * units + j for j in range(units)] for i in indices]
                )
                o_blk.weight.data = blk.weight.data[out_indices].clone()
                if blk.bias is not None:
                    o_blk.bias.data = blk.bias.data[out_indices].clone()

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def groups(self) -> int:
        return self.__groups


class MultiScopicBranch(nn.Sequential, SizeMixin):
    """Branch path of the CNN part of the SOTA model
    from CPSC2019 challenge (entry 0416).

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input tensor.
    scopes : Sequence[Sequence[int]]
        Scopes (in terms of `dilation`) for the convolutional layers,
        each sequence of int is for one branch.
    num_filters : Sequence[int] or Sequence[Sequence[int]]
        Number of filters for the convolutional layers.
        If is sequence of int, then convolutionaly layers
        in one branch will have the same number of filters.
    filter_lengths : Sequence[int] or Sequence[Sequence[int]]
        Filter length (kernel size) of the convolutional layers.
        If is sequence of int, then convolutionaly layers
        in one branch will have the same filter length.
    subsample_lengths : int or Sequence[int]
        Subsample length (stride) of the convolutional layers.
        If is sequence of int, then convolutionaly layers
        in one branch will have the same subsample length.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    config : dict
        Other hyper-parameters, including
        dropout, activation choices, weight initializer, etc.

    """

    __name__ = "MultiScopicBranch"

    def __init__(
        self,
        in_channels: int,
        scopes: Sequence[Sequence[int]],
        num_filters: Union[Sequence[int], Sequence[Sequence[int]]],
        filter_lengths: Union[Sequence[int], Sequence[Sequence[int]]],
        subsample_lengths: Union[int, Sequence[int]],
        groups: int = 1,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__scopes = scopes
        self.__num_blocks = len(self.__scopes)
        self.__num_filters = num_filters
        assert len(self.__num_filters) == self.__num_blocks, (
            f"`scopes` indicates {self.__num_blocks} `MultiScopicBasicBlock`s, "
            f"while `num_filters` indicates {len(self.__num_filters)}"
        )
        self.__filter_lengths = filter_lengths
        assert len(self.__filter_lengths) == self.__num_blocks, (
            f"`scopes` indicates {self.__num_blocks} `MultiScopicBasicBlock`s, "
            f"while `filter_lengths` indicates {len(self.__filter_lengths)}"
        )
        if isinstance(subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(subsample_lengths, self.__num_blocks)
            )
        else:
            self.__subsample_lengths = filter_lengths
            assert len(self.__subsample_lengths) == self.__num_blocks, (
                f"`scopes` indicates {self.__num_blocks} `MultiScopicBasicBlock`s, "
                f"while `subsample_lengths` indicates {len(self.__subsample_lengths)}"
            )
        self.__groups = groups
        self.config = CFG(deepcopy(config))

        block_in_channels = self.__in_channels
        for idx in range(self.__num_blocks):
            self.add_module(
                f"block_{idx}",
                MultiScopicBasicBlock(
                    in_channels=block_in_channels,
                    scopes=self.__scopes[idx],
                    num_filters=self.__num_filters[idx],
                    filter_lengths=self.__filter_lengths[idx],
                    subsample_length=self.__subsample_lengths[idx],
                    groups=self.__groups,
                    dropout=self.config.dropouts[idx],
                    **(self.config.block),
                ),
            )
            block_in_channels = self.__num_filters[idx]

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the branch."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    def _assign_weights_lead_wise(
        self, other: "MultiScopicBranch", indices: Sequence[int]
    ) -> None:
        """Assign weights lead-wise.

        This method is used to assign weights from a model with
        a superset of the current model's leads to the current model.

        Parameters
        ----------
        other : MultiScopicBranch
            The model with a superset of the current model's leads.
        indices : Sequence[int]
            The indices of the leads of the current model in the
            superset model.

        Returns
        -------
        None

        """
        for blk, o_blk in zip(self, other):
            blk._assign_weights_lead_wise(o_blk, indices)

    @property
    def in_channels(self) -> int:
        return self.__in_channels


class MultiScopicCNN(nn.Module, SizeMixin, CitationMixin):
    """CNN part of the SOTA model from CPSC2019 challenge (entry 0416).

    This architecture is a multi-branch CNN with multi-scopic convolutions,
    proposed by the winning team of the CPSC2019 challenge and described in
    [1]_. The multi-scopic convolutions are implemented via different dilations.
    Similar architectures can be found in the model DeepLabv3 [2]_.

    Parameters
    ----------
    in_channels : int
        Number of channels (leads) in the input signal tensor.
    config: dict
        Other hyper-parameters of the Module, ref. corr. config file.
        Key word arguments that must be set:

            - scopes: sequence of sequences of sequences of int,
              scopes (in terms of dilation) of each convolution.
            - num_filters: sequence of sequences (of int or of sequences of int),
              number of filters of the convolutional layers,
              with granularity to each block of each branch,
              or to each convolution of each block of each branch.
            - filter_lengths: sequence of sequences (of int or of sequences of int),
              filter length(s) (kernel size(s)) of the convolutions,
              with granularity to each block of each branch,
              or to each convolution of each block of each branch.
            - subsample_lengths: sequence of int or sequence of sequences of int,
              subsampling length(s) (ratio(s)) of all blocks,
              with granularity to each branch or to each block of each branch,
              each subsamples after the last convolution of each block.
            - dropouts: sequence of int or sequence of sequences of int,
              dropout rates of all blocks,
              with granularity to each branch or to each block of each branch,
              each dropouts at the last of each block.
            - groups: int,
              connection pattern (of channels) of the inputs and outputs.
            - block: dict,
              other parameters that can be set for the building blocks.

        For a full list of configurable parameters, ref. corr. config file.

    References
    ----------
    .. [1] Cai, Wenjie, and Danqin Hu. "QRS complex detection using novel deep learning neural networks." IEEE Access (2020).
    .. [2] Chen, Liang-Chieh, et al. "Rethinking atrous convolution for semantic image segmentation."
          Proceedings of the IEEE conference on computer vision and pattern recognition. 2018.

    """

    __name__ = "MultiScopicCNN"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))
        self.__scopes = self.config.scopes
        self.__num_branches = len(self.__scopes)

        self.branches = nn.ModuleDict()
        for idx in range(self.__num_branches):
            self.branches[f"branch_{idx}"] = MultiScopicBranch(
                in_channels=self.__in_channels,
                scopes=self.__scopes[idx],
                num_filters=self.config.num_filters[idx],
                filter_lengths=self.config.filter_lengths[idx],
                subsample_lengths=self.config.subsample_lengths[idx],
                groups=self.config.groups,
                dropouts=self.config.dropouts[idx],
                block=self.config.block,  # a dict
            )

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        branch_out = OrderedDict()
        for idx in range(self.__num_branches):
            key = f"branch_{idx}"
            branch_out[key] = self.branches[key].forward(input)
        # NOTE: ideally, for lead-wise manner, output of each branch should be
        # split into `self.config.groups` groups,
        # channels from the same group from different branches should be first concatenated,
        # and then concatenate the concatenated groups.
        output = torch.cat(
            [branch_out[f"branch_{idx}"] for idx in range(self.__num_branches)],
            dim=1,  # along channels
        )
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the Module.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            The batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the Module.

        """
        out_channels = 0
        for idx in range(self.__num_branches):
            key = f"branch_{idx}"
            _, _branch_oc, _seq_len = self.branches[key].compute_output_shape(
                seq_len, batch_size
            )
            out_channels += _branch_oc
        output_shape = (batch_size, out_channels, _seq_len)
        return output_shape

    def assign_weights_lead_wise(
        self, other: "MultiScopicCNN", indices: Sequence[int]
    ) -> None:
        """Assign weights to the `other` :class:`MultiScopicCNN`
        module in the lead-wise manner

        Parameters
        ----------
        other : MultiScopicCNN
            The other :class:`MultiScopicCNN` module.
        indices : Sequence[int]
            Indices of the :class:`MultiScopicCNN` modules to assign weights.

        Examples
        --------
        >>> from copy import deepcopy
        >>> import torch
        >>> from torch_ecg.models import ECG_CRNN
        >>> from torch_ecg.model_configs import ECG_CRNN_CONFIG
        >>> from torch_ecg.utils.misc import list_sum
        >>> # we create models using 12-lead ECGs and using reduced 6-lead ECGs
        >>> indices = [0, 1, 2, 3, 4, 10]  # chosen randomly, no special meaning
        >>> # chose the lead-wise models
        >>> lead_12_config = deepcopy(ECG_CRNN_CONFIG)
        >>> lead_12_config.cnn.name = "multi_scopic_leadwise"
        >>> lead_6_config = deepcopy(ECG_CRNN_CONFIG)
        >>> lead_6_config.cnn.name = "multi_scopic_leadwise"
        >>> # adjust groups and numbers of filters
        >>> lead_6_config.cnn.multi_scopic_leadwise.groups = 6
        >>> # numbers of filters should be proportional to numbers of groups
        >>> lead_6_config.cnn.multi_scopic_leadwise.num_filters = (
                np.array([[192, 384, 768], [192, 384, 768], [192, 384, 768]]) / 2
            ).astype(int).tolist()
        >>> # we assume that model12 is a trained model on 12-lead ECGs
        >>> model12 = ECG_CRNN(["AF", "PVC", "NSR"], 12, lead_12_config)
        >>> model6 = ECG_CRNN(["AF", "PVC", "NSR"], 6, lead_6_config)
        >>> model12.eval()
        >>> model6.eval()
        >>> # we create tensor12, tensor6 to check the correctness of the assignment of the weights
        >>> tensor12 = torch.zeros(1, 12, 200)  # batch, leads, seq_len
        >>> tensor6 = torch.randn(1, 6, 200)
        >>> # we make tensor12 has identical values as tensor6 at the given leads, and let the other leads have zero values
        >>> tensor12[:, indices, :] = tensor6
        >>> b = "branch_0"  # similarly for other branches
        >>> _, output_shape_12, _ = model12.cnn.branches[b].compute_output_shape()
        >>> _, output_shape_6, _ = model6.cnn.branches[b].compute_output_shape()
        >>> units = output_shape_12 // 12
        >>> out_indices = list_sum([[i * units + j for j in range(units)] for i in indices])
        >>> (model6.cnn.branches[b](tensor6) == model12.cnn.branches[b](tensor12)[:, out_indices, :]).all()
        tensor(False)  # different feature maps
        >>> # here, we assign weights from model12 that correspond to the given leads to model6
        >>> model12.cnn.assign_weights_lead_wise(model6.cnn, indices)
        >>> (model6.cnn.branches[b](tensor6) == model12.cnn.branches[b](tensor12)[:, out_indices, :]).all()
        tensor(True)  # identical feature maps

        """
        assert (
            self.in_channels == self.config.groups
        ), "the current model is not lead-wise"
        assert (
            other.in_channels == other.config.groups
        ), "the other model is not lead-wise"
        assert other.in_channels == len(
            indices
        ), "the length of indices should match the number of channels of the other model"
        assert (
            np.array(self.config.num_filters) * other.in_channels
            == np.array(other.config.num_filters) * self.in_channels
        ).all(), "the number of filters of the two models should be proportional to the number of channels of the other model"
        for idx in range(self.num_branches):
            for b, ob in zip(
                self.branches[f"branch_{idx}"], other.branches[f"branch_{idx}"]
            ):
                b._assign_weights_lead_wise(ob, indices)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def num_branches(self) -> int:
        return self.__num_branches

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1109/access.2020.2997473"]))



"""
"""

from abc import ABC, abstractmethod
from typing import Any, List, Optional, Sequence, Tuple

import numpy as np
import torch.nn as nn
from torch import Tensor

from ..utils.misc import ReprMixin, add_docstring
from ..cfg import DEFAULTS


__all__ = [
    "Augmenter",
]


_augmenter_forward_doc = """Forward method of the augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor, optional
            Batched labels of the ECGs.
        *extra_tensors : Sequence[torch.Tensor], optional
            Extra tensors to be augmented, e.g. masks for custom loss functions, etc.
        **kwargs: dict, optional
            Additional keyword arguments to be passed to the augmenters.

        Returns
        -------
        Sequence[torch.Tensor]
            The augmented ECGs, labels, and optional extra tensors.

        """


class Augmenter(ReprMixin, nn.Module, ABC):
    """Base class for augmenters.

    An Augmentor performs data augmentation on the input ECGs,
    labels, and optional extra tensors.

    """

    __name__ = "Augmentor"

    @add_docstring(_augmenter_forward_doc)
    @abstractmethod
    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor] = None,
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        raise NotImplementedError

    # def __call__(self, sig:Tensor, label:Optional[Tensor]=None, *extra_tensors:Sequence[Tensor], **kwargs:Any) -> Tuple[Tensor, ...]:
    #     """
    #     alias of `self.generate`
    #     """
    #     return self.generate(sig, label, *extra_tensors, **kwargs)

    def get_indices(
        self, prob: float, pop_size: int, scale_ratio: float = 0.1
    ) -> List[int]:
        """Get a list of indices to be selected.

        A random list of indices in the range ``[0, pop_size-1]``
        is generated, with the probability of each index to be selected.

        Parameters
        ----------
        prob : float
            The probability of each index to be selected.
        pop_size : int
            The size of the population.
        scale_ratio : float, default 0.1
            Scale ratio of std of the normal distribution to the population size.

        Returns
        -------
        indices : List[int],
            A list of indices.

        TODO
        ----
        Add parameter `min_dist` so that
        any 2 selected indices are at least `min_dist` apart.

        """
        k = DEFAULTS.RNG.normal(pop_size * prob, scale_ratio * pop_size)
        # print(pop_size * prob, scale_ratio*pop_size)
        k = int(round(np.clip(k, 0, pop_size)))
        indices = DEFAULTS.RNG_sample(list(range(pop_size)), k).tolist()
        return indices



"""
"""

import torch
import pytest

from torch_ecg.cfg import CFG
from torch_ecg.preprocessors import (
    PreprocManager,
    BandPass,
    BaselineRemove,
    Normalize,
    MinMaxNormalize,
    NaiveNormalize,
    ZScoreNormalize,
    Resample,
)


test_sig = torch.randn(2, 12, 8000)


class DummyPreProcessor(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        return sig


def test_preproc_manager() -> None:
    ppm = PreprocManager(random=True)
    assert ppm.random
    assert ppm.empty
    sig = test_sig.clone()
    sig = ppm(sig)

    ppm.add_(DummyPreProcessor())
    assert not ppm.empty
    ppm.add_(BandPass(fs=500))
    ppm.add_(BaselineRemove(fs=500))
    ppm.add_(Normalize(method="min-max"))
    ppm.add_(Resample(fs=300, dst_fs=500), pos=0)

    sig = test_sig.clone()
    sig = ppm(sig)

    config = CFG(
        random=False,
        bandpass={"fs": 500},
        normalize={"method": "min-max"},
        resample={"fs": 500, "dst_fs": 300},
        baseline_remove={"fs": 500},
        xxx={"fs": 500},  # ignored by `from_config`
    )
    ppm = PreprocManager.from_config(config)
    assert not ppm.random

    sig = test_sig.clone()
    sig = ppm(sig)

    ppm.rearrange(
        new_ordering=[
            "resample",
            "bandpass",
            "baseline_remove",
            "normalize",
        ]
    )

    ppm.random = True
    with pytest.warns(
        RuntimeWarning, match="The preprocessors are applied in random order"
    ):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
                "normalize",
            ]
        )
    ppm.random = False

    with pytest.raises(AssertionError, match="Duplicate preprocessor names"):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
                "normalize",
                "bandpass",
            ]
        )
    with pytest.raises(AssertionError, match="Number of preprocessors mismatch"):
        ppm.rearrange(
            new_ordering=[
                "bandpass",
                "baseline_remove",
                "resample",
            ]
        )

    with pytest.warns(RuntimeWarning, match="No preprocessors added to the manager"):
        ppm = PreprocManager.from_config({"random": True})

    with pytest.warns(RuntimeWarning, match="No preprocessors added to the manager"):
        ppm = PreprocManager.from_config({"bandpass": False})

    del ppm, sig


def test_bandpass() -> None:
    bp = BandPass(fs=500)
    sig = test_sig.clone()
    sig = bp(sig)

    bp = BandPass(fs=500, lowcut=0, highcut=40)
    sig = test_sig.clone()
    sig = bp(sig)

    bp = BandPass(fs=500, lowcut=1.5, highcut=None, inplace=False)
    sig = test_sig.clone()
    sig = bp(sig)

    del bp, sig


def test_baseline_remove() -> None:
    br = BaselineRemove(fs=500, inplace=False)
    sig = test_sig.clone()
    sig = br(sig)

    br = BaselineRemove(fs=500, window1=0.3, window2=0.7)
    sig = test_sig.clone()
    sig = br(sig)

    with pytest.warns(
        RuntimeWarning, match="values of `window1` and `window2` are switched"
    ):
        br = BaselineRemove(fs=500, window1=0.7, window2=0.3)

    del br, sig


def test_normalize() -> None:
    norm = Normalize(method="min-max", inplace=False)
    sig = test_sig.clone()
    sig = norm(sig)

    norm = Normalize(method="z-score")
    sig = test_sig.clone()
    sig = norm(sig)

    norm = MinMaxNormalize()
    sig = test_sig.clone()
    sig = norm(sig)

    norm = ZScoreNormalize()
    sig = test_sig.clone()
    sig = norm(sig)

    norm = NaiveNormalize()
    sig = test_sig.clone()
    sig = norm(sig)

    del norm, sig



"""
"""

from copy import deepcopy

from ...cfg import CFG

__all__ = [
    # vanilla densenet
    "densenet_vanilla",
    # custom densenet
    "densenet_leadwise",
]


densenet_vanilla = CFG()
densenet_vanilla.fs = 500
densenet_vanilla.num_layers = [6, 6, 6, 6]
densenet_vanilla.init_num_filters = 64
densenet_vanilla.init_filter_length = 25
densenet_vanilla.init_pool_stride = 2
densenet_vanilla.init_pool_size = 3
densenet_vanilla.init_subsample_mode = "avg"
densenet_vanilla.growth_rates = 16
densenet_vanilla.filter_lengths = 15
densenet_vanilla.subsample_lengths = 2
densenet_vanilla.bn_size = 4
densenet_vanilla.dropout = 0
densenet_vanilla.compression = 0.5
densenet_vanilla.groups = 1
densenet_vanilla.block = CFG(building_block="basic")
densenet_vanilla.transition = CFG()

densenet_leadwise = deepcopy(densenet_vanilla)
densenet_leadwise.init_num_filters = 12 * 8
densenet_leadwise.growth_rates = 12
densenet_leadwise.groups = 12



"""
"""

from typing import Any, List, Optional, Sequence, Tuple

import torch
from torch import Tensor

from .base import Augmenter

__all__ = [
    "LabelSmooth",
]


class LabelSmooth(Augmenter):
    """Label smoothing augmentation.

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of the ECGs to be augmented.
    smoothing : float, default 0.1
        The smoothing factor.
    prob : float, default 0.5
        Probability of applying label smoothing.
    inplace : bool, default True
        If True, the input tensor will be modified inplace.
    **kwargs : dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        ls = LabelSmooth()
        label = torch.randint(0, 2, (32, 26), dtype=torch.float32)
        _, label = ls(None, label)

    """

    __name__ = "LabelSmooth"

    def __init__(
        self,
        fs: Optional[int] = None,
        smoothing: float = 0.1,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.smoothing = smoothing
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.inplace = inplace

    def forward(
        self,
        sig: Optional[Tensor],
        label: Tensor,
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward method to perform label smoothing.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
            Not used, but kept for compatibility with other augmenters.
        label : torch.Tensor
            The input label tensor,
            of shape ``(batch_size, n_classes)``
            or ``(batch_size, seq_len, n_classes)``.
        extra_tensors : Sequence[torch.Tensor], optional
            Not used, but kept for consistency with other augmenters.
        **kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The input ECG tensor, unchanged.
        label : torch.Tensor
            The output label tensor
            of shape ``(batch_size, n_classes)``
            or ``(batch_size, seq_len, n_classes)``.
        extra_tensors : Sequence[torch.Tensor], optional
            Unchanged extra tensors.

        """
        if not self.inplace:
            label = label.clone()
        if self.prob == 0 or self.smoothing == 0:
            return (sig, label, *extra_tensors)
        n_classes = label.shape[-1]
        batch_size = label.shape[0]
        eps = self.smoothing / max(1, n_classes)
        indices = self.get_indices(prob=self.prob, pop_size=batch_size)
        # print(f"indices = {indices}, len(indices) = {len(indices)}")
        label[indices, ...] = (1 - self.smoothing) * label[
            indices, ...
        ] + torch.full_like(label[indices, ...], eps)
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "smoothing",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
"""

from copy import deepcopy

from ....cfg import CFG, DEFAULTS


__all__ = [
    "MITDBTrainCfg",
]


_NAME = "mitdb"


MITDBTrainCfg = CFG()

MITDBTrainCfg.db_dir = None
MITDBTrainCfg.log_dir = DEFAULTS.log_dir / _NAME
MITDBTrainCfg.model_dir = DEFAULTS.model_dir / _NAME
MITDBTrainCfg.checkpoints = DEFAULTS.checkpoints / _NAME
MITDBTrainCfg.log_dir.mkdir(parents=True, exist_ok=True)
MITDBTrainCfg.model_dir.mkdir(parents=True, exist_ok=True)
MITDBTrainCfg.checkpoints.mkdir(parents=True, exist_ok=True)
# MITDBTrainCfg.test_data_dir = _BASE_DIR / "working_dir" / "sample_data"

MITDBTrainCfg.fs = 360
MITDBTrainCfg.n_leads = 2  # or 1
MITDBTrainCfg.torch_dtype = DEFAULTS.DTYPE.TORCH
MITDBTrainCfg.np_dtype = DEFAULTS.DTYPE.NP

MITDBTrainCfg.beat_types = [
    "N",
    "L",
    "R",
    "V",
    "/",
    "A",
    "f",
    "F",
    "j",
    "a",
    "E",
    "J",
    "Q",
]  # "e", "S" not included, since each of them has only one record
MITDBTrainCfg.rhythm_types = [
    "AFIB",
    "AFL",
    "B",
    "IVR",
    "N",
    "NOD",
    "P",
    "SVTA",
    "T",
    "VT",
]  # "AB", "BII", "PREX", "SBR", "VFL" not included, since each of them has only one record


# common confis for all training tasks
MITDBTrainCfg.data_format = "channel_first"

MITDBTrainCfg.keep_checkpoint_max = 20

MITDBTrainCfg.debug = True

# least distance of an valid R peak to two ends of ECG signals
MITDBTrainCfg.rpeaks_dist2border = int(0.5 * MITDBTrainCfg.fs)  # 0.5s
MITDBTrainCfg.qrs_mask_bias = int(0.075 * MITDBTrainCfg.fs)  # bias to rpeaks

# configs of signal preprocessing
MITDBTrainCfg.normalize = CFG(
    method="z-score",
    per_channel=True,
    mean=0.0,
    std=1.0,
)
# frequency band of the filter to apply, should be chosen very carefully
# MITDBTrainCfg.bandpass = None
MITDBTrainCfg.bandpass = CFG(
    lowcut=0.5,
    highcut=45,
    filter_type="fir",
    filter_order=int(0.3 * MITDBTrainCfg.fs),
)

# configs of data aumentation
# MITDBTrainCfg.label_smooth = CFG(
#     prob=0.8,
#     smoothing=0.1,
# )
MITDBTrainCfg.label_smooth = False
MITDBTrainCfg.random_masking = False
MITDBTrainCfg.stretch_compress = False  # stretch or compress in time axis
# MITDBTrainCfg.mixup = CFG(
#     prob=0.6,
#     alpha=0.3,
# )
MITDBTrainCfg.random_flip = CFG(
    per_channel=True,
    prob=[0.4, 0.5],
)

# TODO: explore and add more data augmentations

# configs of training epochs, batch, etc.
MITDBTrainCfg.n_epochs = 30
MITDBTrainCfg.batch_size = 64
MITDBTrainCfg.train_ratio = 0.8

# configs of optimizers and lr_schedulers
MITDBTrainCfg.optimizer = "adamw_amsgrad"  # "sgd", "adam", "adamw"
MITDBTrainCfg.momentum = 0.949  # default values for corresponding PyTorch optimizers
MITDBTrainCfg.betas = (
    0.9,
    0.999,
)  # default values for corresponding PyTorch optimizers
MITDBTrainCfg.decay = 1e-2  # default values for corresponding PyTorch optimizers

MITDBTrainCfg.learning_rate = 1e-4  # 1e-3
MITDBTrainCfg.lr = MITDBTrainCfg.learning_rate

MITDBTrainCfg.lr_scheduler = (
    "one_cycle"  # "one_cycle", "plateau", "burn_in", "step", None
)
MITDBTrainCfg.lr_step_size = 50
MITDBTrainCfg.lr_gamma = 0.1
MITDBTrainCfg.max_lr = 2e-3  # for "one_cycle" scheduler, to adjust via expriments

MITDBTrainCfg.early_stopping = CFG()  # early stopping according to challenge metric
MITDBTrainCfg.early_stopping.min_delta = 0.001  # should be non-negative
MITDBTrainCfg.early_stopping.patience = 8

# configs of loss function
# "MaskedBCEWithLogitsLoss", "BCEWithLogitsWithClassWeightLoss"  # "BCELoss"
# MITDBTrainCfg.loss = "AsymmetricLoss"
# MITDBTrainCfg.loss_kw = CFG(gamma_pos=0, gamma_neg=1, implementation="deep-psp")
MITDBTrainCfg.loss = "MaskedBCEWithLogitsLoss"
MITDBTrainCfg.loss_kw = CFG()
MITDBTrainCfg.flooding_level = 0.0  # flooding performed if positive

MITDBTrainCfg.log_step = 40

# tasks of training
MITDBTrainCfg.tasks = [
    "beat_classification",
    "qrs_detection",
    "rhythm_segmentation",
    "af_event",  # segmentation of AF events
    "rr_lstm",
]

# configs of model selection
# "resnet_leadwise", "multi_scopic_leadwise", "vgg16", "resnet", "vgg16_leadwise", "cpsc", "cpsc_leadwise", etc.

for t in MITDBTrainCfg.tasks:
    MITDBTrainCfg[t] = CFG()


MITDBTrainCfg.beat_classification.final_model_name = None
MITDBTrainCfg.beat_classification.model_name = "crnn"
MITDBTrainCfg.beat_classification.cnn_name = "resnet_nature_comm_bottle_neck_se"
MITDBTrainCfg.beat_classification.rnn_name = "none"
MITDBTrainCfg.beat_classification.attn_name = "se"  # "none", "se", "gc", "nl"
MITDBTrainCfg.beat_classification.winL = int(0.8 * MITDBTrainCfg.fs)
MITDBTrainCfg.beat_classification.winR = int(1.2 * MITDBTrainCfg.fs)
MITDBTrainCfg.beat_classification.input_len = (
    MITDBTrainCfg.beat_classification.winL + MITDBTrainCfg.beat_classification.winR
)
MITDBTrainCfg.beat_classification.classes = deepcopy(MITDBTrainCfg.beat_types)
MITDBTrainCfg.beat_classification.class_map = {
    k: i for i, k in enumerate(MITDBTrainCfg.beat_classification.classes)
}
MITDBTrainCfg.beat_classification.monitor = (
    "f1_measure"  # monitor for determining the best model
)
MITDBTrainCfg.beat_classification.loss = "AsymmetricLoss"  # "BCEWithLogitsLoss"
MITDBTrainCfg.beat_classification.loss_kw = CFG(
    gamma_pos=0, gamma_neg=0.5, implementation="deep-psp"
)

MITDBTrainCfg.qrs_detection.final_model_name = None
MITDBTrainCfg.qrs_detection.model_name = "seq_lab"  # "unet"
MITDBTrainCfg.qrs_detection.reduction = 1
MITDBTrainCfg.qrs_detection.cnn_name = "multi_scopic"
MITDBTrainCfg.qrs_detection.rnn_name = "lstm"  # "none", "lstm"
MITDBTrainCfg.qrs_detection.attn_name = "se"  # "none", "se", "gc", "nl"
MITDBTrainCfg.qrs_detection.input_len = int(30 * MITDBTrainCfg.fs)
MITDBTrainCfg.qrs_detection.overlap_len = int(15 * MITDBTrainCfg.fs)
MITDBTrainCfg.qrs_detection.critical_overlap_len = int(25 * MITDBTrainCfg.fs)
MITDBTrainCfg.qrs_detection.classes = [
    "N",
]
MITDBTrainCfg.qrs_detection.monitor = (
    "qrs_score"  # monitor for determining the best model
)
MITDBTrainCfg.qrs_detection.loss = "BCEWithLogitsLoss"  # "AsymmetricLoss"
MITDBTrainCfg.qrs_detection.loss_kw = CFG()

MITDBTrainCfg.rr_lstm.final_model_name = None
MITDBTrainCfg.rr_lstm.model_name = "lstm"  # "lstm", "lstm_crf"
MITDBTrainCfg.rr_lstm.input_len = 30  # number of rr intervals ( number of rpeaks - 1)
MITDBTrainCfg.rr_lstm.overlap_len = 15  # number of rr intervals ( number of rpeaks - 1)
MITDBTrainCfg.rr_lstm.critical_overlap_len = (
    25  # number of rr intervals ( number of rpeaks - 1)
)
MITDBTrainCfg.rr_lstm.classes = [
    "AFIB",
]
MITDBTrainCfg.rr_lstm.monitor = "neg_masked_bce"  # "rr_score", "neg_masked_bce"  # monitor for determining the best model
MITDBTrainCfg.rr_lstm.loss = "MaskedBCEWithLogitsLoss"
MITDBTrainCfg.rr_lstm.loss_kw = CFG()

for t in ["rhythm_segmentation", "af_event"]:  # segmentation of AF events
    MITDBTrainCfg[t].final_model_name = None
    MITDBTrainCfg[t].model_name = "seq_lab"  # "unet"
    MITDBTrainCfg[t].reduction = 1
    MITDBTrainCfg[t].cnn_name = "multi_scopic"
    MITDBTrainCfg[t].rnn_name = "lstm"  # "none", "lstm"
    MITDBTrainCfg[t].attn_name = "se"  # "none", "se", "gc", "nl"
    MITDBTrainCfg[t].input_len = int(30 * MITDBTrainCfg.fs)
    MITDBTrainCfg[t].overlap_len = int(15 * MITDBTrainCfg.fs)
    MITDBTrainCfg[t].critical_overlap_len = int(25 * MITDBTrainCfg.fs)
    MITDBTrainCfg[
        t
    ].monitor = "neg_masked_bce"  # "main_score", "neg_masked_bce"  # monitor for determining the best model
    # MITDBTrainCfg[t].loss = "AsymmetricLoss" # "MaskedBCEWithLogitsLoss"
    # MITDBTrainCfg[t].loss_kw = CFG(gamma_pos=0, gamma_neg=1, implementation="deep-psp")
    MITDBTrainCfg[t].loss = "MaskedBCEWithLogitsLoss"  # "MaskedBCEWithLogitsLoss"
    MITDBTrainCfg[t].loss_kw = CFG()

MITDBTrainCfg.rhythm_segmentation.classes = deepcopy(MITDBTrainCfg.rhythm_types)
MITDBTrainCfg.rhythm_segmentation.class_map = {
    k: i for i, k in enumerate(MITDBTrainCfg.rhythm_segmentation.classes)
}
MITDBTrainCfg.af_event.classes = ["AFIB"]



# -*- coding: utf-8 -*-

from collections import defaultdict
from numbers import Real
from pathlib import Path
from typing import Any, Optional, Union, Sequence, List, Dict

import numpy as np
import pandas as pd
import scipy.signal as SS
import wfdb
from tqdm.auto import tqdm

from ...cfg import DEFAULTS
from ...utils import (
    add_docstring,
    get_record_list_recursive3,
    generalized_intervals_intersection,
)
from ..base import PhysioNetDataBase, DataBaseInfo, PSGDataBaseMixin


__all__ = [
    "CINC2018",
]


_CINC2018_INFO = DataBaseInfo(
    title="""
    You Snooze You Win - The PhysioNet Computing in Cardiology Challenge 2018
    """,
    about="""
    1. includes 1,985 subjects, partitioned into balanced training (n = 994), and test sets (n = 989)
    2. signals include

        electrocardiogram (ECG),
        electroencephalography (EEG),
        electrooculography (EOG),
        electromyography (EMG),
        electrocardiology (EKG),
        oxygen saturation (SaO2),
        etc.

    3. frequency of all signal channels is 200 Hz
    4. units of signals:

        mV for ECG, EEG, EOG, EMG, EKG
        percentage for SaO2

    5. six sleep stages were annotated in 30 second contiguous intervals:

        wakefulness,
        stage 1,
        stage 2,
        stage 3,
        rapid eye movement (REM),
        undefined

    6. annotated arousals were classified as either of the following:

        spontaneous arousals,
        respiratory effort related arousals (RERA),
        bruxisms,
        hypoventilations,
        hypopneas,
        apneas (central, obstructive and mixed),
        vocalizations,
        snores,
        periodic leg movements,
        Cheyne-Stokes breathing,
        partial airway obstructions

    7. Webpage of the database on PhysioNet [1]_.
    """,
    usage=[
        "sleep stage",
        "sleep apnea",
    ],
    references=[
        "https://physionet.org/content/challenge-2018/",
    ],
    doi=[
        "10.22489/CinC.2018.049",
        "10.13026/6phb-r450",
    ],
)


@add_docstring(_CINC2018_INFO.format_database_docstring(), mode="prepend")
class CINC2018(PhysioNetDataBase, PSGDataBaseMixin):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "CINC2018"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[str] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="challenge-2018",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 200
        self._subset = kwargs.get("subset", "training")
        self.rec_ext = "mat"
        self.ann_ext = "arousal"

        # fmt: off
        self.sleep_stage_names = ["W", "R", "N1", "N2", "N3"]
        self.arousal_types = [
            "arousal_bruxism", "arousal_noise", "arousal_plm", "arousal_rera", "arousal_snore", "arousal_spontaneous",
            "resp_centralapnea", "resp_cheynestokesbreath", "resp_hypopnea", "resp_hypoventilation",
            "resp_mixedapnea", "resp_obstructiveapnea", "resp_partialobstructive",
        ]
        # fmt: on

        self.training_rec_pattern = "^tr\\d{2}\\-\\d{4}.mat$"
        self.test_rec_pattern = "^te\\d{2}\\-\\d{4}.mat$"
        self.training_records = []
        self.test_records = []
        self._all_records = []
        self._df_records = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        self._df_records = pd.DataFrame()
        records = get_record_list_recursive3(
            self.db_dir,
            {"training": self.training_rec_pattern, "test": self.test_rec_pattern},
            relative=False,
        )
        for k in records:
            df_tmp = pd.DataFrame(sorted(records[k]), columns=["path"])
            df_tmp["subset"] = k
            self._df_records = pd.concat(
                [self._df_records, df_tmp], axis=0, ignore_index=True
            )
        self._df_records["record"] = self._df_records["path"].apply(
            lambda x: Path(x).stem
        )
        self._df_records["subject_id"] = self._df_records["record"].apply(
            self.get_subject_id
        )
        self._df_records.set_index("record", inplace=True)

        self._df_records["fs"] = None
        self._df_records["siglen"] = None
        self._df_records["available_signals"] = None
        with tqdm(
            self._df_records.iterrows(),
            total=len(self._df_records),
            mininterval=1.0,
            desc="Loading metadata",
            disable=self.verbose < 1,
        ) as pbar:
            for idx, row in pbar:
                header = wfdb.rdheader(row["path"])
                self._df_records.at[idx, "fs"] = header.fs
                self._df_records.at[idx, "siglen"] = header.sig_len
                self._df_records.at[idx, "available_signals"] = header.sig_name

        if self._subset is not None:
            self._df_records = self._df_records[
                self._df_records["subset"] == self._subset
            ]

        if self._subsample is not None:
            if self._subset is None:
                df_tmp = pd.DataFrame(columns=self._df_records.columns)
                for k in records:
                    size = int(round(self._subsample * len(records[k])))
                    if size > 0:
                        df_tmp = pd.concat(
                            [
                                df_tmp,
                                self._df_records[
                                    self._df_records["subset"] == k
                                ].sample(
                                    size, random_state=DEFAULTS.SEED, replace=False
                                ),
                            ],
                            axis=0,
                            ignore_index=True,
                        )
                if len(df_tmp) == 0:
                    size = min(
                        len(self._df_records),
                        max(1, int(round(self._subsample * len(self._df_records)))),
                    )
                    df_tmp = self._df_records.sample(
                        size, random_state=DEFAULTS.SEED, replace=False
                    )
                del self._df_records
                self._df_records = df_tmp.copy()
                del df_tmp
            else:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                if size > 0:
                    self._df_records = self._df_records.sample(
                        size, random_state=DEFAULTS.SEED, replace=False
                    )

        self._all_records = self._df_records.index.tolist()
        self.training_records = self._df_records[
            self._df_records["subset"] == "training"
        ].index.tolist()
        self.test_records = self._df_records[
            self._df_records["subset"] == "test"
        ].index.tolist()

    def get_subject_id(self, rec: str) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        head = "2018"
        mid = rec[2:4]
        tail = rec[-4:]
        pid = int(head + mid + tail)
        return pid

    def set_subset(self, subset: Union[str, None]) -> None:
        """Set the subset of the database to use."""
        assert subset in [
            "training",
            "test",
            None,
        ], """`subset` must be in ``["training", "test", None]``."""
        self._subset = subset
        self._ls_rec()

    def get_available_signals(self, rec: Union[str, int]) -> List[str]:
        """Get the available signals of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        signals : List[str]
            Names of available signal of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return self._df_records.at[rec, "available_signals"]

    def get_fs(self, rec: Union[str, int]) -> int:
        """Get the sampling frequency of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        fs : int
            Sampling frequency of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return self._df_records.at[rec, "fs"]

    def get_siglen(self, rec: Union[str, int]) -> int:
        """Get the length of the signal of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        siglen : int
            Length of the signal of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return self._df_records.at[rec, "siglen"]

    def load_psg_data(
        self,
        rec: Union[str, int],
        channel: Optional[Union[str, Sequence[str]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        physical: bool = True,
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """Load PSG data of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        channel : str, optional
            Nname of the channel of PSG data.
            If is None, all channels will be returned.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        data_format: str, default "channel_first".
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain") which is valid
            only when only one `channel` is passed.
        physical : bool, default True
            If True, the data will be converted to physical units,
            otherwise, the data will be in digital units.
        fs : numbers.Real, optional
            Sampling frequency of the output signal.
            If not None, the loaded data will be resampled to this frequency,
            otherwise, the original sampling frequency will be used.

        Returns
        -------
        numpy.ndarray
            PSG data corr. to the given `channel` of the record.

        """
        available_signals = self.get_available_signals(rec)
        chn = available_signals if channel is None else channel
        if isinstance(chn, str):
            chn = [chn]
        assert set(chn).issubset(
            set(available_signals)
        ), f"`channel` should be one of `{available_signals}`, but got `{chn}`"

        allowed_data_format = [
            "channel_first",
            "lead_first",
            "channel_last",
            "lead_last",
            "flat",
            "plain",
        ]
        assert (
            data_format.lower() in allowed_data_format
        ), f"`data_format` should be one of `{allowed_data_format}`, but got `{data_format}`"
        if len(chn) > 1:
            assert data_format.lower() in [
                "channel_first",
                "lead_first",
                "channel_last",
                "lead_last",
            ], (
                "`data_format` should be one of "
                "`['channel_first', 'lead_first', 'channel_last', 'lead_last']` "
                f"when the passed number of `channel` is larger than 1, but got `{data_format}`"
            )

        frp = str(self.get_absolute_path(rec))
        wfdb_header = wfdb.rdheader(frp)
        sampfrom = max(0, sampfrom or 0)
        sampto = min(sampto or wfdb_header.sig_len, wfdb_header.sig_len)
        wfdb_rec = wfdb.rdrecord(
            frp, sampfrom=sampfrom, sampto=sampto, channel_names=chn, physical=physical
        )

        ret_data = wfdb_rec.p_signal.T if physical else wfdb_rec.d_signal.T

        if fs is not None and fs != wfdb_header.fs:
            ret_data = SS.resample_poly(ret_data, fs, wfdb_header.fs, axis=-1)

        if data_format.lower() in ["channel_last", "lead_last"]:
            ret_data = ret_data.T
        elif data_format.lower() in ["flat", "plain"]:
            ret_data = ret_data.flatten()

        return ret_data

    def load_data(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: Union[str, type(None)] = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """Load ECG data of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or Sequence[str] or Sequence[int], optional
            The leads of the ECG data to load.
            None or "all" for all leads.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain") which is valid only when `leads` is a single lead
        units : str or None, default "mV"
            Units of the output signal, can also be "μV" (aliases "uV", "muV").
            None for digital data, without digital-to-physical conversion.
        fs : numbers.Real, optional
            Sampling frequency of the output signal.
            If not None, the loaded data will be resampled to this frequency,
            otherwise, the original sampling frequency will be used.

        Returns
        -------
        numpy.ndarray
            The ECG data loaded from the record,
            with given `units` and `data_format`.

        """
        available_signals = self.get_available_signals(rec)
        assert (
            "ECG" in available_signals
        ), f"the record `{rec}` does not have ECG signal"
        allowed_units = ["mv", "uv", "μv", "muv"]
        assert (
            units is None or units.lower() in allowed_units
        ), f"`units` should be one of `{allowed_units}` or None, but got `{units}`"
        data = self.load_psg_data(
            rec=rec,
            channel="ECG",
            sampfrom=sampfrom,
            sampto=sampto,
            data_format=data_format,
            physical=units is not None,
            fs=fs,
        )
        if units.lower() in ["μv", "uv", "muv"]:
            data = 1000 * data
        return data

    @add_docstring(load_data.__doc__)
    def load_ecg_data(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: Union[str, type(None)] = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """alias of `load_data`"""
        return self.load_data(
            rec=rec,
            sampfrom=sampfrom,
            sampto=sampto,
            data_format=data_format,
            units=units,
            fs=fs,
        )

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
    ) -> Dict[str, Dict[str, List[List[int]]]]:
        """Load sleep stage and arousal annotations of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the corresponding PSG data.
        sampto : int, optional
            End index of the corresponding PSG data.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        dict
            A dictionary with keys "sleep_stages" and "arousals",
            each of which is a dictionary with keys of sleep stages and arousals,
            and values of lists of lists of start and
            end indices of the sleep stages and arousals.

        """
        frp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(frp, extension=self.ann_ext)

        sleep_stages = defaultdict(list)
        arousals = defaultdict(list)
        current_sleep_stage = None
        current_sleep_stage_start = None
        for aux_note, sample in zip(wfdb_ann.aux_note, wfdb_ann.sample.tolist()):
            if aux_note in self.sleep_stage_names:
                if current_sleep_stage is not None:
                    sleep_stages[current_sleep_stage].append(
                        [current_sleep_stage_start, sample]
                    )
                current_sleep_stage = aux_note
                current_sleep_stage_start = sample
            else:
                if "(" in aux_note:
                    current_arousal_start = sample
                else:
                    arousals[aux_note.strip(")")].append(
                        [current_arousal_start, sample]
                    )
        siglen = self.get_siglen(rec)
        if current_sleep_stage_start < siglen:
            sleep_stages[current_sleep_stage].append(
                [current_sleep_stage_start, siglen]
            )
        sampfrom = max(0, sampfrom or 0)
        sampto = min(sampto or siglen, siglen)
        sleep_stages = {
            k: generalized_intervals_intersection(
                v, [[sampfrom, sampto]], drop_degenerate=True
            )
            for k, v in sleep_stages.items()
        }
        sleep_stages = {k: v for k, v in sleep_stages.items() if len(v) > 0}
        arousals = {
            k: generalized_intervals_intersection(
                v, [[sampfrom, sampto]], drop_degenerate=True
            )
            for k, v in arousals.items()
        }
        arousals = {k: v for k, v in arousals.items() if len(v) > 0}
        if not keep_original:
            sleep_stages = {
                k: [[s - sampfrom, e - sampfrom] for s, e in v]
                for k, v in sleep_stages.items()
            }
            arousals = {
                k: [[s - sampfrom, e - sampfrom] for s, e in v]
                for k, v in arousals.items()
            }
        return {
            "sleep_stages": sleep_stages,
            "arousals": arousals,
        }

    def load_sleep_stages_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
    ) -> Dict[str, List[List[int]]]:
        """Load sleep stage annotations of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the corresponding PSG data.
        sampto : int, optional
            End index of the corresponding PSG data.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        dict
            A dictionary with keys of sleep stages and
            values of lists of lists of start and
            end indices of the sleep stages.

        """
        return self.load_ann(
            rec=rec,
            sampfrom=sampfrom,
            sampto=sampto,
            keep_original=keep_original,
        )["sleep_stages"]

    def load_arousals_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
    ) -> Dict[str, List[List[int]]]:
        """Load arousal annotations of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the corresponding PSG data.
        sampto : int, optional
            End index of the corresponding PSG data.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        dict
            A dictionary with keys of arousals and
            values of lists of lists of start and
            end indices of the arousals.

        """
        return self.load_ann(
            rec=rec,
            sampfrom=sampfrom,
            sampto=sampto,
            keep_original=keep_original,
        )["arousals"]

    def plot(self) -> None:
        """NOT implemented yet."""
        raise NotImplementedError

    def plot_ann(self, rec: Union[str, int]) -> tuple:
        """Plot the sleep stage and arousal annotations of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object.
        ax : matplotlib.axes.Axes
            The axes object.

        TODO
        ----
        Plot arousals events.

        """
        ann = self.load_ann(rec)
        sleep_stages = ann["sleep_stages"]
        arousals = ann["arousals"]
        stage_mask = self.sleep_stage_intervals_to_mask(sleep_stages)
        fig, ax = self.plot_hypnogram(stage_mask)
        # TODO: plot arousals events
        return fig, ax

    @property
    def database_info(self) -> DataBaseInfo:
        return _CINC2018_INFO



# -*- coding: utf-8 -*-
"""
Base classes for datasets from different sources:

    - PhysioNet
    - NSRR
    - CPSC
    - Other databases

Remarks
-------
1. For whole-dataset visualizing: http://zzz.bwh.harvard.edu/luna/vignettes/dataplots/
2. Visualizing using UMAP: http://zzz.bwh.harvard.edu/luna/vignettes/nsrr-umap/

"""

import logging
import os
import posixpath
import pprint
import re
import textwrap
import time
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from string import punctuation
from typing import Any, List, Optional, Union, Sequence, Dict
from numbers import Real

import numpy as np
import pandas as pd
import requests
import scipy.signal as SS
import wfdb
from pyedflib import EdfReader

from ..cfg import CFG, DEFAULTS, _DATA_CACHE
from ..utils import ecg_arrhythmia_knowledge as EAK  # noqa : F401
from ..utils.download import http_get
from ..utils.misc import (
    dict_to_str,
    get_record_list_recursive,
    init_logger,
    ReprMixin,
    CitationMixin,
)
from .aux_data import get_physionet_dbs


__all__ = [
    "WFDB_Beat_Annotations",
    "WFDB_Non_Beat_Annotations",
    "WFDB_Rhythm_Annotations",
    "PhysioNetDataBase",
    "NSRRDataBase",
    "CPSCDataBase",
    "DEFAULT_FIG_SIZE_PER_SEC",
    "BeatAnn",
    "DataBaseInfo",
    "PSGDataBaseMixin",
]


WFDB_Beat_Annotations = {
    "N": "Normal beat",
    "L": "Left bundle branch block beat",
    "R": "Right bundle branch block beat",
    "B": "Bundle branch block beat (unspecified)",
    "A": "Atrial premature beat",
    "a": "Aberrated atrial premature beat",
    "J": "Nodal (junctional) premature beat",
    "S": "Supraventricular premature or ectopic beat (atrial or nodal)",
    "V": "Premature ventricular contraction",
    "r": "R-on-T premature ventricular contraction",
    "F": "Fusion of ventricular and normal beat",
    "e": "Atrial escape beat",
    "j": "Nodal (junctional) escape beat",
    "n": "Supraventricular escape beat (atrial or nodal)",
    "E": "Ventricular escape beat",
    "/": "Paced beat",
    "f": "Fusion of paced and normal beat",
    "Q": "Unclassifiable beat",
    "?": "Beat not classified during learning",
}

WFDB_Non_Beat_Annotations = {
    "[": "Start of ventricular flutter/fibrillation",
    "!": "Ventricular flutter wave",
    "]": "End of ventricular flutter/fibrillation",
    "x": "Non-conducted P-wave (blocked APC)",
    "(": "Waveform onset",
    ")": "Waveform end",
    "p": "Peak of P-wave",
    "t": "Peak of T-wave",
    "u": "Peak of U-wave",
    "`": "PQ junction",
    "'": "J-point",
    "^": "(Non-captured) pacemaker artifact",
    "|": "Isolated QRS-like artifact",
    "~": "Change in signal quality",
    "+": "Rhythm change",
    "s": "ST segment change",
    "T": "T-wave change",
    "*": "Systole",
    "D": "Diastole",
    "=": "Measurement annotation",
    '"': "Comment annotation",
    "@": "Link to external data",
}

WFDB_Rhythm_Annotations = {
    "(AB": "Atrial bigeminy",
    "(AFIB": "Atrial fibrillation",
    "(AFL": "Atrial flutter",
    "(B": "Ventricular bigeminy",
    "(BII": "2° heart block",
    "(IVR": "Idioventricular rhythm",
    "(N": "Normal sinus rhythm",
    "(NOD": "Nodal (A-V junctional) rhythm",
    "(P": "Paced rhythm",
    "(PREX": "Pre-excitation (WPW)",
    "(SBR": "Sinus bradycardia",
    "(SVTA": "Supraventricular tachyarrhythmia",
    "(T": "Ventricular trigeminy",
    "(VFL": "Ventricular flutter",
    "(VT": "Ventricular tachycardia",
}


class _DataBase(ReprMixin, ABC):
    """Universal abstract base class for all databases.

    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments

    """

    def __init__(
        self,
        db_name: str,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        self.db_name = db_name
        if db_dir is None:
            db_dir = _DATA_CACHE / db_name
            warnings.warn(
                f"`db_dir` is not specified, "
                f"using default `{db_dir}` as the storage path",
                RuntimeWarning,
            )
        self.db_dir = Path(db_dir).expanduser().resolve().absolute()
        if not self.db_dir.exists():
            self.db_dir.mkdir(parents=True, exist_ok=True)
            warnings.warn(
                f"`{self.db_dir}` does not exist. It is now created. "
                "Please check if it is set correctly. "
                "Or if you may want to download the database into this folder, "
                "please use the `download()` method.",
                RuntimeWarning,
            )
        self.working_dir = (
            Path(working_dir or DEFAULTS.working_dir).expanduser().resolve().absolute()
            / self.db_name
        )
        self.working_dir.mkdir(parents=True, exist_ok=True)

        self.logger = kwargs.get("logger", None)
        if self.logger is None:
            self.logger = init_logger(
                log_dir=False,
                suffix=self.__class__.__name__,
                verbose=verbose,
            )
        else:
            assert isinstance(
                self.logger, logging.Logger
            ), "logger must be a `logging.Logger` instance"

        self.data_ext = None
        self.ann_ext = None
        self.header_ext = "hea"
        self.verbose = verbose
        self._df_records = pd.DataFrame()
        self._all_records = None

        self._subsample = kwargs.get("subsample", None)
        assert (
            self._subsample is None or 0 < self._subsample <= 1
        ), f"`subsample` must be in (0, 1], but got `{self._subsample}`"

    @abstractmethod
    def _ls_rec(self) -> None:
        """Find all records in the database."""
        raise NotImplementedError

    @abstractmethod
    def load_data(self, rec: Union[str, int], **kwargs) -> Any:
        """Load data from the record."""
        raise NotImplementedError

    @abstractmethod
    def load_ann(self, rec: Union[str, int], **kwargs) -> Any:
        """Load annotations of the record.

        NOTE that the records might have several annotation files.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def database_info(self) -> "DataBaseInfo":
        """The :class:`DataBaseInfo` object of the database."""
        raise NotImplementedError

    def get_citation(
        self, format: Optional[str] = None, style: Optional[str] = None
    ) -> None:
        """Get the citations of the papers related to the database.

        Parameters
        ----------
        lookup : bool, default True
            Whether to lookup the citation from the DOI or not.
        format : str, optional
            Format of the final output
            If specified, the default format ("bib") will be overrided.
        style : str, optional
            Style of the final output.
            If specified, the default style ("apa") will be overrided.
            Valid only when `format` is ``"text"``.
        print_result : bool, default False
            Whether to print the final output
            instead of returning it or not.

        Returns
        -------
        None

        """
        self.database_info.get_citation(
            lookup=True, format=format, style=style, timeout=10.0, print_result=True
        )

    def _auto_infer_units(self, sig: np.ndarray, sig_type: str = "ECG") -> str:
        """Automatically infer the units of the signal.

        It is assumed that `sig` is not raw signal, but with baseline removed.

        Parameters
        ----------
        sig : ndarray
            The signal to infer its units.
        sig_type : str, default "ECG"
            Type of the signal, case insensitive.

        Returns
        -------
        units : {"μV", "mV"}
            Units of the signal.

        """
        if sig_type.lower() == "ecg":
            _MAX_mV = 20  # 20mV, seldom an ECG device has range larger than this value
            max_val = np.max(np.abs(sig))
            if max_val > _MAX_mV:
                units = "μV"
            else:
                units = "mV"
        else:
            raise NotImplementedError(f"not implemented for {sig_type}")
        return units

    @property
    def all_records(self) -> List[str]:
        if self._all_records is None:
            self._ls_rec()
        return self._all_records

    def get_absolute_path(
        self, rec: Union[str, int], extension: Optional[str] = None
    ) -> Path:
        """Get the absolute path of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        extension : str, optional
            Extension of the file.

        Returns
        -------
        path : pathlib.Path
            Absolute path of the file.

        """
        if isinstance(rec, int):
            rec = self[rec]
        path = self._df_records.loc[rec].path
        if extension is not None:
            path = path.with_suffix(
                extension if extension.startswith(".") else f".{extension}"
            )
        return path

    def _normalize_leads(
        self,
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        all_leads: Optional[Sequence[str]] = None,
        numeric: bool = False,
    ) -> List[Union[str, int]]:
        """Normalize the leads to a list of standard lead names.

        Parameters
        ----------
        leads : str or int or List[str] or List[int], optional
            the (names of) leads to normalize
        all_leads : list of str, optional
            All leads of the records in the database,
            If is None, the database class should have attribute `all_leads`,
            and `self.all_leads` will be used.
        numeric : bool, default False
            If True, indices of the leads will be returned
            instead of lead names.

        Returns
        -------
        leads : List[str] or List[int]
            The normalized leads

        """
        if all_leads is None:
            assert hasattr(
                self, "all_leads"
            ), "If `all_leads` is not specified, the database class should have attribute `all_leads`!"
            all_leads = self.all_leads
        err_msg = (
            f"`leads` should be a subset of {all_leads} or non-negative integers "
            f"less than {len(all_leads)}, but got {leads}"
        )
        if leads is None or (isinstance(leads, str) and leads.lower() == "all"):
            _leads = all_leads
        elif isinstance(leads, str):
            _leads = [leads]
        elif isinstance(leads, int):
            assert len(all_leads) > leads >= 0, err_msg
            _leads = [all_leads[leads]]
        else:
            try:
                _leads = [ld if isinstance(ld, str) else all_leads[ld] for ld in leads]
            except Exception:
                raise AssertionError(err_msg)
        assert set(_leads).issubset(all_leads), err_msg
        if numeric:
            _leads = [all_leads.index(ld) for ld in _leads]
        return _leads

    @classmethod
    def get_arrhythmia_knowledge(cls, arrhythmias: Union[str, List[str]]) -> None:
        """Knowledge about ECG features of specific arrhythmias.

        Parameters
        ----------
        arrhythmias : str or List[str]
            The arrhythmia(s) to check,
            in abbreviations or in SNOMEDCTCode.

        Returns
        -------
        None

        """
        if isinstance(arrhythmias, str):
            d = [arrhythmias]
        else:
            d = arrhythmias
        for idx, item in enumerate(d):
            print(dict_to_str(eval(f"EAK.{item}")))
            if idx < len(d) - 1:
                print("*" * 110)

    def extra_repr_keys(self) -> List[str]:
        return [
            "db_name",
            "db_dir",
        ]

    @property
    @abstractmethod
    def url(self) -> Union[str, List[str]]:
        """URL(s) for downloading the database."""
        raise NotImplementedError

    def __len__(self) -> int:
        return len(self.all_records)

    def __getitem__(self, index: int) -> str:
        return self.all_records[index]


class PhysioNetDataBase(_DataBase):
    """Base class for readers for PhysioNet database.

    PhysioNet is a large repository of freely available biomedical signals,
    including ECG, EEG, EMG, and other signals.
    The website is [#phy_website]_.

    Parameters
    ----------
    db_name : str
        Name of the database.
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If is None, `wfdb` will fetch data from PhysioNet.
    working_dir : str or pathlib.Path, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Verbosity level for logging.
    kwargs : dict, optional
        Auxilliary key word arguments.

    References
    ----------
    .. [#phy_website] https://www.physionet.org/

    """

    def __init__(
        self,
        db_name: str,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name=db_name,
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        # `self.fs` for those with single signal source, e.g. ECG,
        # for those with multiple signal sources like PSG,
        # self.fs is default to the frequency of ECG if ECG applicable
        self.fs = None
        self._all_records = None
        self._version = None
        self._url_compressed = None

        self.df_all_db_info = get_physionet_dbs()

        if self.verbose > 2:
            self.df_all_db_info = (
                pd.DataFrame(
                    wfdb.get_dbs(),
                    columns=[
                        "db_name",
                        "db_description",
                    ],
                )
                .drop_duplicates()
                .reset_index(drop=True)
            )

    def _ls_rec(self, db_name: Optional[str] = None, local: bool = True) -> None:
        """
        Find all records (relative path without file extension),
        and save into some private attributes for further use.

        Parameters
        ----------
        db_name : str, optional
            Name of the database for using :meth:`wfdb.get_record_list`.
            If is None, :attr:`self.db_name` will be used.
        local : bool, default True
            If True, search records in local storage,
            prior using :meth:`wfdb.get_record_list`.

        Returns
        -------
        None

        """
        empty_warning_msg = (
            "No records found in the database! "
            "Please check if path to the database is correct. "
            "Or you can try to download the database first using the `download` method."
        )
        if local:
            self._ls_rec_local()
            if len(self._df_records) == 0:
                warnings.warn(empty_warning_msg, RuntimeWarning)
            return
        try:
            self._df_records = pd.DataFrame()
            self._df_records["record"] = wfdb.get_record_list(db_name or self.db_name)
            self._df_records["path"] = self._df_records["record"].apply(
                lambda x: (self.db_dir / x).resolve()
            )
            # keep only the records that exist in `self.db_dir`
            # NOTE
            # 1. data files might be in some subdirectories of `self.db_dir`
            # 2. `wfdb.get_record_list` will return records without file extension
            self._df_records = self._df_records[
                self._df_records["path"].apply(
                    lambda x: len(x.parent.glob(f"{x.name}.*")) > 0
                )
            ]
            # if no record found,
            # search locally and recursively inside `self.db_dir`
            if len(self._df_records) == 0:
                return self._ls_rec_local()
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.name
            )  # remove relative path, leaving only the record name
            self._df_records.set_index("record", inplace=True)
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self.logger.debug(
                    f"subsample `{size}` records from `{len(self._df_records)}`"
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )
            self._all_records = self._df_records.index.tolist()
        except Exception:
            self._ls_rec_local()
        if len(self._df_records) == 0:
            warnings.warn(empty_warning_msg, RuntimeWarning)

    def _ls_rec_local(self) -> None:
        """Find all records in :attr:`self.db_dir`."""
        record_list_fp = self.db_dir / "RECORDS"
        self._df_records = pd.DataFrame()
        if record_list_fp.is_file():
            self._df_records["record"] = [
                item
                for item in record_list_fp.read_text().splitlines()
                if len(item) > 0
            ]
            if len(self._df_records) > 0:
                if self._subsample is not None:
                    size = min(
                        len(self._df_records),
                        max(1, int(round(self._subsample * len(self._df_records)))),
                    )
                    self.logger.debug(
                        f"subsample `{size}` records from `{len(self._df_records)}`"
                    )
                    self._df_records = self._df_records.sample(
                        n=size, random_state=DEFAULTS.SEED, replace=False
                    )
                self._df_records["path"] = self._df_records["record"].apply(
                    lambda x: (self.db_dir / x).resolve()
                )
                self._df_records = self._df_records[
                    self._df_records["path"].apply(lambda x: x.is_file())
                ]
                self._df_records["record"] = self._df_records["path"].apply(
                    lambda x: x.name
                )

        if len(self._df_records) == 0:
            print(
                "Please wait patiently to let the reader find "
                "all records of the database from local storage..."
            )
            start = time.time()
            self._df_records["path"] = get_record_list_recursive(
                self.db_dir, self.data_ext, relative=False
            )
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self.logger.debug(
                    f"subsample `{size}` records from `{len(self._df_records)}`"
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )
            self._df_records["path"] = self._df_records["path"].apply(lambda x: Path(x))
            self.logger.info(f"Done in {time.time() - start:.3f} seconds!")
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.name
            )
        self._df_records.set_index("record", inplace=True)
        self._all_records = self._df_records.index.values.tolist()

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        raise NotImplementedError

    def load_data(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: Union[str, type(None)] = "mV",
        fs: Optional[Real] = None,
    ) -> np.ndarray:
        """
        Load physical (converted from digital) ECG data,
        which is more understandable for humans;
        or load digital signal directly.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or Sequence[str] or Sequence[int], optional
            The leads of the ECG data to load.
            None or "all" for all leads.
        sampfrom : int, optional
            Start index of the data to be loaded.
        sampto : int, optional
            End index of the data to be loaded.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first"), or
            "flat" (alias "plain") which is valid only when `leads` is a single lead
        units : str or None, default "mV"
            Units of the output signal, can also be "μV" (aliases "uV", "muV").
            None for digital data, without digital-to-physical conversion.
        fs : numbers.Real, optional
            Sampling frequency of the output signal.
            If not None, the loaded data will be resampled to this frequency,
            otherwise, the original sampling frequency will be used.

        Returns
        -------
        numpy.ndarray
            The ECG data loaded from the record,
            with given `units` and `data_format`.

        """
        fp = str(self.get_absolute_path(rec))

        if hasattr(self, "all_leads"):
            all_leads = self.all_leads
        else:
            all_leads = wfdb.rdheader(fp).sig_name
        _leads = self._normalize_leads(leads, all_leads, numeric=False)

        allowed_data_format = [
            "channel_first",
            "lead_first",
            "channel_last",
            "lead_last",
            "flat",
            "plain",
        ]
        assert (
            data_format.lower() in allowed_data_format
        ), f"`data_format` should be one of `{allowed_data_format}`, but got `{data_format}`"
        if len(_leads) > 1:
            assert data_format.lower() in [
                "channel_first",
                "lead_first",
                "channel_last",
                "lead_last",
            ], (
                "`data_format` should be one of `['channel_first', 'lead_first', 'channel_last', 'lead_last']` "
                f"when the passed number of `leads` is larger than 1, but got `{data_format}`"
            )

        allowed_units = ["mv", "uv", "μv", "muv"]
        assert (
            units is None or units.lower() in allowed_units
        ), f"`units` should be one of `{allowed_units}` or None, but got `{units}`"

        rdrecord_kwargs = dict(
            sampfrom=sampfrom or 0,
            sampto=sampto,
            physical=units is not None,
            return_res=DEFAULTS.DTYPE.INT,
            channels=[all_leads.index(ld) for ld in _leads],
        )  # use `channels` instead of `channel_names` since there're exceptional cases where `channel_names` has duplicates
        wfdb_rec = wfdb.rdrecord(fp, **rdrecord_kwargs)

        # p_signal or d_signal is in the format of "lead_last", and with units in "mV"
        if units is None:
            data = wfdb_rec.d_signal
        elif units.lower() == "mv":
            data = wfdb_rec.p_signal
        elif units.lower() in ["μv", "uv", "muv"]:
            data = 1000 * wfdb_rec.p_signal

        if fs is not None and hasattr(self, "fs") and fs != self.fs:
            data = SS.resample_poly(data, fs, self.fs, axis=0).astype(data.dtype)

        if data_format.lower() in ["channel_first", "lead_first"]:
            data = data.T
        elif data_format.lower() in ["flat", "plain"]:
            data = data.flatten()

        return data

    def helper(self, items: Union[List[str], str, type(None)] = None) -> None:
        """Print corr. meanings of symbols belonging to `items`.

        More details can be found
        in the PhysioNet WFDB annotation manual [#ann_man]_.

        Parameters
        ----------
        items : str or List[str], optional
            Items to print.
            If is None, then a comprehensive printing
            of meanings of all symbols will be performed.

        Returns
        -------
        None

        References
        ----------
        .. [#ann_man] https://archive.physionet.org/physiobank/annotations.shtml

        """
        attrs = vars(self)
        methods = [
            func
            for func in dir(self)
            if callable(getattr(self, func))
            and not (func.startswith("__") and func.endswith("__"))
        ]

        beat_annotations = deepcopy(WFDB_Beat_Annotations)
        non_beat_annotations = deepcopy(WFDB_Non_Beat_Annotations)
        rhythm_annotations = deepcopy(WFDB_Rhythm_Annotations)

        all_annotations = [
            beat_annotations,
            non_beat_annotations,
            rhythm_annotations,
        ]

        summary_items = [
            "beat",
            "non-beat",
            "rhythm",
        ]

        if items is None:
            _items = [
                "attributes",
                "methods",
                "beat",
                "non-beat",
                "rhythm",
            ]
        elif isinstance(items, str):
            _items = [items]
        else:
            _items = items

        pp = pprint.PrettyPrinter(indent=4)

        if "attributes" in _items:
            print("--- helpler - attributes ---")
            pp.pprint(attrs)
        if "methods" in _items:
            print("--- helpler - methods ---")
            pp.pprint(methods)
        if "beat" in _items:
            print("--- helpler - beat ---")
            pp.pprint(beat_annotations)
        if "non-beat" in _items:
            print("--- helpler - non-beat ---")
            pp.pprint(non_beat_annotations)
        if "rhythm" in _items:
            print("--- helpler - rhythm ---")
            pp.pprint(rhythm_annotations)

        for k in _items:
            if k in summary_items:
                continue
            for a in all_annotations:
                if k in a.keys() or "(" + k in a.keys():
                    try:
                        print(f"`{k.split('(')[1]}` stands for `{a[k]}`")
                    except IndexError:
                        try:
                            print(f"`{k}` stands for `{a[k]}`")
                        except KeyError:
                            print(f"`{k}` stands for `{a['('+k]}`")

    def get_file_download_url(self, file_name: Union[str, Path]) -> str:
        """Get the download url of the file.

        Parameters
        ----------
        file_name : str or pathlib.Path
            Name of the file,
            e.g. "data/001a.dat", "training/tr03-0005/tr03-0005.mat", etc.

        Returns
        -------
        url : str
            URL of the file to be downloaded.

        """
        url = posixpath.join(
            wfdb.io.download.PN_INDEX_URL,
            self.db_name,
            self.version,
            file_name,
        )
        return url

    @property
    def version(self) -> str:
        """Version of the database."""
        if self._version is not None:
            return self._version
        try:
            self._version = wfdb.io.record.get_version(self.db_name)
        except Exception:
            warnings.warn(
                "Cannot get the version number from PhysioNet! Defaults to '1.0.0'",
                RuntimeWarning,
            )
            self._version = "1.0.0"
        return self._version

    @property
    def webpage(self) -> str:
        """URL of the database webpage"""
        return posixpath.join(
            wfdb.io.download.PN_CONTENT_URL, f"{self.db_name}/{self.version}"
        )

    @property
    def url(self) -> str:
        """URL of the database index page for downloading."""
        return posixpath.join(
            wfdb.io.download.PN_INDEX_URL, f"{self.db_name}/{self.version}"
        )

    @property
    def url_(self) -> Union[str, type(None)]:
        """URL of the compressed database file for downloading."""
        if self._url_compressed is not None:
            return self._url_compressed
        domain = "https://physionet.org/static/published-projects/"
        punct = re.sub("[\\-:]", "", punctuation)
        try:
            db_desc = self.df_all_db_info[
                self.df_all_db_info["db_name"] == self.db_name
            ].iloc[0]["db_description"]
        except IndexError:
            self.logger.info(
                f"\042{self.db_name}\042 is not in the database list hosted at PhysioNet!"
            )
            return None
        db_desc = re.sub(f"[{punct}]+", "", db_desc).lower()
        db_desc = re.sub("[\\s:]+", "-", db_desc)
        url = posixpath.join(domain, f"{self.db_name}/{db_desc}-{self.version}.zip")
        if requests.head(url).headers.get("Content-Type") == "application/zip":
            self._url_compressed = url
        else:
            new_url = posixpath.join(
                wfdb.io.download.PN_INDEX_URL, f"{self.db_name}/get-zip/{self.version}"
            )
            print(f"{url} is not available, try {new_url} instead")
        return self._url_compressed

    def download(self, compressed: bool = True) -> None:
        """Download the database from PhysioNet."""
        if compressed:
            if self.url_ is not None:
                http_get(self.url_, self.db_dir, extract=True)
                self._ls_rec()
                return
            else:
                self.logger.info(
                    "No compressed database available! Downloading the uncompressed version..."
                )
        wfdb.dl_database(
            self.db_name,
            self.db_dir,
            keep_subdirs=True,
            overwrite=False,
        )
        self._ls_rec()


class NSRRDataBase(_DataBase):
    """Base class for readers for the NSRR database.

    For a full list of available databases, and their descriptions,
    please visit the NSRR database webpage [1]_.

    Parameters
    ----------
    db_name : str
        Name of the database.
    db_dir : str or pathlib.Path, optional
        Local storage path of the database.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Verbosity level for logging.
    kwargs : dict, optional
        Auxilliary key word arguments.

    References
    ----------
    .. [1] https://sleepdata.org/

    """

    def __init__(
        self,
        db_name: str,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[str] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name=db_name,
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = None
        self._all_records = None
        self.file_opened = None

        all_dbs = [
            [
                "shhs",
                "Multi-cohort study focused on sleep-disordered breathing and cardiovascular outcomes",
            ],
            ["mesa", ""],
            ["oya", ""],
            [
                "chat",
                "Multi-center randomized trial comparing early adenotonsillectomy to "
                "watchful waiting plus supportive care",
            ],
            [
                "heartbeat",
                "Multi-center Phase II randomized controlled trial that evaluates the effects "
                "of supplemental nocturnal oxygen or Positive Airway Pressure (PAP) therapy",
            ],
            # more to be added
        ]
        self.df_all_db_info = pd.DataFrame(
            {
                "db_name": [item[0] for item in all_dbs],
                "db_description": [item[1] for item in all_dbs],
            }
        )
        self.kwargs = kwargs

    def safe_edf_file_operation(
        self,
        operation: str = "close",
        full_file_path: Optional[Union[str, Path]] = None,
    ) -> None:
        """Safe IO operation for edf file.

        Parameters
        ----------
        operation : {"open", "close"}, optional
            Operation name, by default "close".
        full_file_path : str or pathlib.Path, optional
            Path of the file which contains the data.
            If is None, default path will be used.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the operation is not supported.

        """
        if operation == "open":
            if self.file_opened is not None:
                self.file_opened._close()
            self.file_opened = EdfReader(str(full_file_path))
        elif operation == "close":
            if self.file_opened is not None:
                self.file_opened._close()
                self.file_opened = None
        else:
            raise ValueError("Illegal operation")

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        raise NotImplementedError

    def show_rec_stats(self, rec: Union[str, int]) -> None:
        """Print the statistics about the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        None

        """
        raise NotImplementedError

    def helper(self, items: Union[List[str], str, type(None)] = None) -> None:
        """Print corr. meanings of symbols belonging to `items`.

        Parameters
        ----------
        items : str or List[str], optional
            Items to print.
            If is None, then a comprehensive printing
            of meanings of all symbols will be performed.

        Returns
        -------
        None

        """
        pp = pprint.PrettyPrinter(indent=4)

        attrs = vars(self)
        methods = [
            func
            for func in dir(self)
            if callable(getattr(self, func))
            and not (func.startswith("__") and func.endswith("__"))
        ]

        if items is None:
            _items = [
                "attributes",
                "methods",
            ]
        elif isinstance(items, str):
            _items = [items]
        else:
            _items = items

        pp = pprint.PrettyPrinter(indent=4)

        if "attributes" in _items:
            print("--- helpler - attributes ---")
            pp.pprint(attrs)
        if "methods" in _items:
            print("--- helpler - methods ---")
            pp.pprint(methods)


class CPSCDataBase(_DataBase):
    """Base class for readers for the CPSC database.

    Parameters
    ----------
    db_name : str
        Name of the database.
    db_dir : str or pathlib.Path, optional
        Local storage path of the database.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Verbosity level for logging.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    def __init__(
        self,
        db_name: str,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[str] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name=db_name,
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = None
        self._all_records = None

        self.kwargs = kwargs

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        raise NotImplementedError

    def helper(self, items: Union[List[str], str, type(None)] = None) -> None:
        """Print corr. meanings of symbols belonging to `items`.

        Parameters
        ----------
        items : str or List[str], optional
            Items to print.
            If is None, then a comprehensive printing
            of meanings of all symbols will be performed.

        Returns
        -------
        None

        """
        pp = pprint.PrettyPrinter(indent=4)

        attrs = vars(self)
        methods = [
            func
            for func in dir(self)
            if callable(getattr(self, func))
            and not (func.startswith("__") and func.endswith("__"))
        ]

        if items is None:
            _items = [
                "attributes",
                "methods",
            ]
        elif isinstance(items, str):
            _items = [items]
        else:
            _items = items

        pp = pprint.PrettyPrinter(indent=4)

        if "attributes" in _items:
            print("--- helpler - attributes ---")
            pp.pprint(attrs)
        if "methods" in _items:
            print("--- helpler - methods ---")
            pp.pprint(methods)

    def download(self) -> None:
        """Download the database from `self.url`."""
        if isinstance(self.url, str):
            http_get(self.url, self.db_dir, extract=True)
        else:
            for url in self.url:
                http_get(url, self.db_dir, extract=True)
        self._ls_rec()


@dataclass
class DataBaseInfo(CitationMixin):
    """A dataclass to store the information of a database.

    Attributes
    ----------
    title : str
        Title of the database.
    about : str or list of str
        Description of the database.
    usage : list of str
        Potential usages of the database.
    references : list of str
        References of the database.
    note : str or list of str, optional
        Notes of the database.
    issues : str or list of str, optional
        Known issues of the database.
    status : str, optional
        Status of the database.
    doi : str or list of str, optional
        DOI of the paper(s) describing the database.

    """

    title: str
    about: Union[str, Sequence[str]]
    usage: Sequence[str]
    references: Sequence[str]
    note: Optional[Union[str, Sequence[str]]] = None
    issues: Optional[Union[str, Sequence[str]]] = None
    status: Optional[str] = None
    doi: Optional[Union[str, Sequence[str]]] = None

    def format_database_docstring(self, indent: Optional[str] = None) -> str:
        """Format the database docstring from
        the information stored in the dataclass.

        The docstring will use the reStructuredText format.

        Parameters
        ----------
        indent : str, optional
            Indent of the docstring.
            If not specified, then 4 spaces will be used.

        Returns
        -------
        str
            The formatted docstring.

        NOTE
        ----
        An environment variable ``DB_BIB_LOOKUP`` can be set to
        `True` to enable the lookup of the bib entries.

        """
        if indent is None:
            indent = " " * 4
        title = textwrap.dedent(self.title).strip("\n ")
        if isinstance(self.about, str):
            about = "ABOUT\n-----\n" + textwrap.dedent(self.about).strip("\n ")
        else:
            about = ["ABOUT", "-----"] + [
                f"{idx+1}. {line}" for idx, line in enumerate(self.about)
            ]
            about = "\n".join(about)
        if self.note is None:
            # note = "NOTE\n----"
            note = ""
        elif isinstance(self.note, str):
            note = "NOTE\n----\n" + textwrap.dedent(self.note).strip("\n ")
        else:
            note = ["NOTE", "----"] + [
                f"{idx+1}. {line}" for idx, line in enumerate(self.note)
            ]
            note = "\n".join(note)
        if self.issues is None:
            # issues = "Issues\n------"
            issues = ""
        elif isinstance(self.issues, str):
            issues = "Issues\n------\n" + textwrap.dedent(self.issues).strip("\n ")
        else:
            issues = ["Issues", "-" * 6] + [
                f"{idx+1}. {line}" for idx, line in enumerate(self.issues)
            ]
            issues = "\n".join(issues)
        references = ["References", "-" * 10] + [
            # f"""{idx+1}. <a name="ref{idx+1}"></a> {line}"""
            f""".. [{idx+1}] {line}"""
            for idx, line in enumerate(self.references)
        ]
        references = "\n".join(references)
        usage = ["Usage", "------"] + [
            f"{idx+1}. {line}" for idx, line in enumerate(self.usage)
        ]
        usage = "\n".join(usage)

        docstring = textwrap.indent(
            f"""\n{title}\n\n{about}\n\n{note}\n\n{usage}\n\n{issues}\n\n{references}\n""",
            indent,
        )

        if self.status is not None and len(self.status) > 0:
            docstring = f"{self.status}\n\n{docstring}"

        lookup = os.getenv("DB_BIB_LOOKUP", False)
        citation = self.get_citation(lookup=lookup, print_result=False)
        if citation.startswith("@"):
            citation = textwrap.indent(citation, indent)
            citation = textwrap.indent(
                f"""Citation\n--------\n.. code-block:: bibtex\n\n{citation}""", indent
            )
            docstring = f"{docstring}\n\n{citation}\n"
        elif not lookup:
            citation = textwrap.indent(f"""Citation\n--------\n{citation}""", indent)
            docstring = f"{docstring}\n\n{citation}\n"

        return docstring


class PSGDataBaseMixin:
    """A mixin class for PSG databases.

    Contains methods for

        - convertions between sleep stage intervals and sleep stage masks
        - hypnogram plotting

    """

    def sleep_stage_intervals_to_mask(
        self,
        intervals: Dict[str, List[List[int]]],
        fs: Optional[int] = None,
        granularity: int = 30,
        class_map: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        """Convert sleep stage intervals to sleep stage mask.

        Parameters
        ----------
        intervals : dict
            Sleep stage intervals, in the format of dict of list of lists of int.
            Keys are sleep stages and
            values are lists of lists of start and end indices of the sleep stages.
        fs : int, optional
            Sampling frequency corresponding to the sleep stage intervals,
            defaults to the sampling frequency of the database.
        granularity : int, default 30
            Granularity of the sleep stage mask, with units in seconds.
        class_map : dict, optional
            A dictionary mapping sleep stages to integers.
            If the database reader does not have a `sleep_stage_names` attribute,
            this parameter must be provided.

        Returns
        -------
        numpy.ndarray
            Sleep stage mask.

        """
        fs = fs or self.fs
        assert fs is not None and fs > 0, "`fs` must be positive"
        assert granularity > 0, "`granularity` must be positive"
        if not hasattr(self, "sleep_stage_names"):
            assert class_map is not None, "`class_map` must be provided"
        else:
            class_map = class_map or {
                k: len(self.sleep_stage_names) - i - 1
                for i, k in enumerate(self.sleep_stage_names)
            }
        intervals = {
            class_map[k]: [
                [int(round(s / fs / granularity)), int(round(e / fs / granularity))]
                for s, e in v
            ]
            for k, v in intervals.items()
        }
        intervals = {k: [[s, e] for s, e in v if s < e] for k, v in intervals.items()}
        intervals = {k: v for k, v in intervals.items() if len(v) > 0}
        siglen = max([e for v in intervals.values() for s, e in v])
        mask = np.zeros(siglen, dtype=int)
        for k, v in intervals.items():
            for s, e in v:
                mask[s:e] = k
        return mask

    def plot_hypnogram(
        self,
        mask: np.ndarray,
        granularity: int = 30,
        class_map: Optional[Dict[str, int]] = None,
        **kwargs,
    ) -> tuple:
        """Hypnogram visualization.

        Parameters
        ----------
        mask : numpy.ndarray
            Sleep stage mask.
        granularity : int, default 30
            Granularity of the sleep stage mask to be plotted,
            with units in seconds.
        class_map : dict, optional
            A dictionary mapping sleep stages to integers.
            If the database reader does not have a `sleep_stage_names` attribute,
            this parameter must be provided.
        kwargs : dict, optional
            Additional keyword arguments passed to :meth:`matplotlib.pyplot.plot`.

        Returns
        -------
        fig : matplotlib.figure.Figure
            Figure object.
        ax : matplotlib.axes._subplots.AxesSubplot
            Axes object.

        """
        if not hasattr(self, "sleep_stage_names"):
            pass
        else:
            class_map = class_map or {
                k: len(self.sleep_stage_names) - i - 1
                for i, k in enumerate(self.sleep_stage_names)
            }

        if "plt" not in globals():
            import matplotlib.pyplot as plt

        fig_width = (
            len(mask) * granularity / 3600 / 6 * 20
        )  # stardard width is 20 for 6 hours

        fig, ax = plt.subplots(figsize=(fig_width, 4))
        color = kwargs.pop("color", "black")
        ax.plot(mask, color=color, **kwargs)

        # xticks to the format of HH:MM, every half hour
        xticks = np.arange(0, len(mask), 1800 / granularity)
        xticklabels = [
            f"{int(i * granularity / 3600):02d}:{int(i * granularity / 60 % 60):02d}"
            for i in xticks
        ]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, fontsize=14)
        ax.set_xlabel("Time", fontsize=18)
        ax.set_xlim(0, len(mask))
        # yticks to the format of sleep stages
        yticks = sorted(class_map.values())
        yticklabels = [k for k, v in sorted(class_map.items(), key=lambda x: x[1])]
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticklabels, fontsize=14)
        ax.set_ylabel("Sleep Stage", fontsize=18)

        return fig, ax


DEFAULT_FIG_SIZE_PER_SEC = 4.8


@dataclass
class BeatAnn:
    """Dataclass for beat annotation.

    Attributes
    ----------
    index : int
        Index of the beat.
    symbol : str
        Symbol of the beat.

    Properties
    ----------
    name : str
        Name of the beat.

    """

    index: int
    symbol: str

    @property
    def name(self) -> str:
        if self.symbol in WFDB_Beat_Annotations:
            return WFDB_Beat_Annotations[self.symbol]
        return WFDB_Non_Beat_Annotations.get(self.symbol, self.symbol)


# configurations for visualization
_PlotCfg = CFG()
# used only when corr. values are absent
# all values are time bias w.r.t. corr. peaks, with units in ms
_PlotCfg.p_onset = -40
_PlotCfg.p_offset = 40
_PlotCfg.q_onset = -20
_PlotCfg.s_offset = 40
_PlotCfg.qrs_radius = 60
_PlotCfg.t_onset = -100
_PlotCfg.t_offset = 60



"""
Cutmix augmentation, for segmentation tasks.
This technique is very successful in CPSC2021 challenge of paroxysmal AF events detection.
"""

from copy import deepcopy
from numbers import Real
from random import shuffle
from typing import Any, List, Optional, Sequence, Tuple

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter
from ..cfg import DEFAULTS


__all__ = [
    "CutMix",
]


class CutMix(Augmenter):
    """CutMix augmentation.

    CutMix is a data augmentation technique originally proposed in [1]_,
    with official implementation in [2]_, and an unofficial implementation in [3]_.
    This technique was designed for image classification tasks, but it can also be used
    for ECG tasks. This technique was very successful
    in CPSC2021 challenge of paroxysmal AF events detection.

    Parameters
    ----------
    fs: int, optional
        Sampling frequency, by default None.
    num_mix: int, default 1
        Number of mixtures.
    alpha: float, default 0.5
        Beta distribution parameter.
    beta: float, optional
        Beta distribution parameter, by default equal to `alpha`.
    prob: float, default 0.5
        Probability of applying this augmenter.
    inplace: bool, default True
        Whether to perform this augmentation in-place.
    **kwargs: dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        cm = CutMix(prob=0.7)
        sig = torch.randn(32, 12, 5000)
        lb = torch.randint(0, 2, (32, 5000, 2), dtype=torch.float32)  # 2 classes mask
        sig, lb = cm(sig, lb)

    References
    ----------
    .. [1] Yun, S., Han, D., Oh, S. J., Chun, S., Choe, J., & Yoo, Y. (2019).
       CutMix: Regularization strategy to train strong classifiers with localizable features.
       In Proceedings of the IEEE/CVF International Conference on Computer Vision (pp. 6023-6032).
    .. [2] https://github.com/clovaai/CutMix-PyTorch/blob/master/train.py
    .. [3] https://github.com/ildoonet/cutmix/blob/master/cutmix/cutmix.py

    """

    __name__ = "CutMix"

    def __init__(
        self,
        fs: Optional[int] = None,
        num_mix: int = 1,
        alpha: Real = 0.5,
        beta: Optional[Real] = None,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.fs = fs
        self.num_mix = num_mix
        assert (
            isinstance(self.num_mix, int) and self.num_mix > 0
        ), f"`num_mix` must be a positive integer, but got `{self.num_mix}`"
        self.alpha = alpha
        self.beta = beta or self.alpha
        assert (
            self.alpha > 0 and self.beta > 0
        ), f"`alpha` and `beta` must be positive, but got `{self.alpha}` and `{self.beta}`"
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Tensor,
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any,
    ) -> Tuple[Tensor, ...]:
        """Forward method to perform CutMix augmentation.

        Parameters
        ----------
        sig: torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label: torch.Tensor
            Class (one-hot) labels, of shape ``(batch, num_classes)``;
            or segmentation masks, of shape ``(batch, siglen, num_classes)``.
        extra_tensors: Sequence[torch.Tensor], optional
            Other tensors to be augmented, by default None.
        **kwargs: dict, optional
            Additional keyword arguments.

        Returns
        -------
        Tuple[torch.Tensor]
            Augmented tensors.

        """
        assert label.ndim != 1, "`label` should NOT be categorical labels"

        if not self.inplace:
            sig = sig.clone()
            label = label.clone()
            extra_tensors = [t.clone() for t in extra_tensors]
        else:
            extra_tensors = list(extra_tensors)  # make it mutable
        batch, lead, siglen = sig.shape

        for _ in range(self.num_mix):
            indices = np.arange(batch, dtype=int)
            # original indices chosen by probability
            ori = self.get_indices(prob=self.prob, pop_size=batch)
            # permuted indices
            perm = deepcopy(ori)
            shuffle(perm)
            indices[ori] = perm
            indices = torch.from_numpy(indices).long()

            lam = torch.from_numpy(
                # DEFAULTS.RNG.beta(self.alpha, self.beta, size=len(ori)),
                DEFAULTS.RNG.beta(self.alpha, self.beta, size=batch),
            ).to(
                dtype=sig.dtype, device=sig.device
            )  # shape: (batch,)
            intervals = _make_intervals(lam, siglen)

            # perform cutmix in batch
            # set values of sig enclosed by intervals to 0
            mask = torch.ones_like(sig)
            for i, (start, end) in enumerate(intervals):
                mask[i, :, start:end] = 0
            sig = sig * mask + sig[indices] * (1 - mask)

            if label.ndim == 2:
                # one-hot labels, shape (batch, num_classes)
                label = label * lam.view(-1, 1) + label[indices] * (1 - lam.view(-1, 1))
            else:
                # segmentation masks, shape (batch, siglen, num_classes)
                label = label * lam.view(-1, 1, 1) + label[indices] * (
                    1 - lam.view(-1, 1, 1)
                )

            for i, t in enumerate(extra_tensors):
                if t.ndim == 2:
                    extra_tensors[i] = t * lam.view(-1, 1) + t[indices] * (
                        1 - lam.view(-1, 1)
                    )
                else:
                    extra_tensors[i] = t * lam.view(-1, 1, 1) + t[indices] * (
                        1 - lam.view(-1, 1, 1)
                    )

        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "alpha",
            "beta",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()


def _make_intervals(lam: Tensor, siglen: int) -> np.ndarray:
    """Make intervals for cutmix.

    Parameters
    ----------
    lam : torch.Tensor
        Parameter ``lambda`` for cutmix, of shape ``(n,)``.
    siglen : int
        Length of the signal.

    Returns
    -------
    numpy.ndarray
        Intervals for cutmix, of shape ``(n, 2)``.

    """
    _lam = (lam.numpy() * siglen).astype(int)
    intervals = np.zeros((lam.shape[0], 2), dtype=int)
    intervals[:, 0] = np.minimum(
        DEFAULTS.RNG_randint(0, siglen, size=lam.shape[0]), siglen - _lam
    )
    intervals[:, 1] = intervals[:, 0] + _lam
    return intervals



import json
import math
import warnings
from copy import deepcopy
from random import shuffle
from typing import List, Optional, Sequence, Tuple, Any

import numpy as np
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....cfg import CFG
from ....databases import CPSC2019 as CR
from ....utils.misc import ReprMixin


__all__ = [
    "CPSC2019Dataset",
]


class CPSC2019Dataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.CPSC2019` database.

    Parameters
    ----------
    config : dict
        Configurations for the dataset, ref. `CPSC2019TrainCfg`.
        A simple example is as follows:

        .. code-block:: python

            >>> config = deepcopy(CPSC2019TrainCfg)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = CPSC2019Dataset(config, training=True, lazy=False)

    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __name__ = "CPSC2019Dataset"

    def __init__(
        self,
        config: CFG,
        training: bool = True,
        lazy: bool = False,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = CR(db_dir=config.db_dir, **reader_kwargs)
        self.config.db_dir = self.reader.db_dir
        self.training = training
        self.n_classes = 1
        self.lazy = lazy

        self.dtype = self.config.np_dtype

        self.siglen = self.config.input_len  # alias, for simplicity
        self.records = []
        self._train_test_split(
            train_ratio=self.config.train_ratio,
            force_recompute=False,
        )
        self.ppm = PreprocManager.from_config(self.config)

        self.fdr = _FastDataReader(self.reader, self.records, self.config, self.ppm)

        self._signals = None
        self._labels = None
        if not self.lazy:
            self._load_all_data()

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        if self.lazy:
            signal, label = self.fdr[index]
        else:
            signal, label = self._signals[index], self._labels[index]
        return signal, label

    def __len__(self) -> int:
        return len(self.fdr)

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        self._signals, self._labels = [], []
        with tqdm(
            self.fdr,
            desc="loading data",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for sig, lab in pbar:
                self._signals.append(sig)
                self._labels.append(lab)
        self._signals = np.array(self._signals)
        self._labels = np.array(self._labels)

    @property
    def signals(self) -> np.ndarray:
        """Cached signals, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._signals

    @property
    def labels(self) -> np.ndarray:
        """Cached labels, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._labels

    def _train_test_split(
        self, train_ratio: float = 0.8, force_recompute: bool = False
    ) -> List[str]:
        """Perform train-test split.

        Parameters
        ----------
        train_ratio : float, default 0.8
            Ratio of the train set in the whole dataset.
        force_recompute : bool, default False
            If True, the train-test split will be recomputed,
            regardless of the existing ones stored in json files.

        Returns
        -------
        records : List[str]
            List of the records split for training
            or for testing (validation).

        """
        assert 0 < train_ratio < 100
        _train_ratio = train_ratio if train_ratio < 1 else train_ratio / 100
        split_fn = self.reader.db_dir / f"train_test_split_{_train_ratio:.2f}.json"
        if split_fn.is_file() and not force_recompute:
            split_res = json.loads(split_fn.read_text())
            if self.training:
                self.records = split_res["train"]
                shuffle(self.records)
            else:
                self.records = split_res["test"]
            return
        records = deepcopy(self.reader.all_records)
        shuffle(records)
        split_num = int(_train_ratio * len(records))
        train = sorted(records[:split_num])
        test = sorted(records[split_num:])
        split_res = {"train": train, "test": test}
        split_fn.write_text(json.dumps(split_res, ensure_ascii=False))
        if self.training:
            self.records = train
            shuffle(self.records)
        else:
            self.records = test

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        reader: CR,
        records: Sequence[str],
        config: CFG,
        ppm: Optional[PreprocManager] = None,
    ) -> None:
        self.reader = reader
        self.records = records
        self.config = config
        self.ppm = ppm

        self.siglen = self.config.input_len  # alias, for simplicity

    def __len__(self) -> int:
        return len(self.records)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        rec_name = self.records[index]
        values = self.reader.load_data(rec_name, units="mV", data_format="flat")
        rpeaks = self.reader.load_ann(rec_name)
        if self.config.get("recover_length", False):
            reduction = 1
        else:
            reduction = self.config.reduction
        labels = np.zeros((self.siglen // reduction))
        # rpeak indices to mask
        for r in rpeaks:
            if r < self.config.skip_dist or r >= self.siglen - self.config.skip_dist:
                continue
            start_idx = math.floor((r - self.config.bias_thr) / reduction)
            end_idx = math.ceil((r + self.config.bias_thr) / reduction)
            labels[start_idx:end_idx] = 1

        values = values.reshape((self.config.n_leads, self.siglen))
        labels = labels[..., np.newaxis]

        values, _ = self.ppm(values, self.config.fs)

        return values, labels

    def extra_repr_keys(self) -> List[str]:
        return [
            "reader",
            "ppm",
        ]



"""
"""

from abc import ABC, abstractmethod
from typing import Any, Union, Sequence, Set, Dict

import numpy as np
import pandas as pd

from ..cfg import CFG
from ..utils.misc import add_docstring
from ..utils.utils_data import ECGWaveFormNames
from .metrics import (
    ClassificationMetrics,
    RPeaksDetectionMetrics,
    WaveDelineationMetrics,
)


__all__ = [
    "BaseOutput",
    "ClassificationOutput",
    "MultiLabelClassificationOutput",
    "SequenceTaggingOutput",
    "SequenceLabellingOutput",
    "WaveDelineationOutput",
    "RPeaksDetectionOutput",
]


_KNOWN_ISSUES = """
    NOTE
    ----
    Known issues:

        - fields of type `dict` are not well supported due to the limitations of the base class `CFG`, for example

        .. code-block:: python

        {}
    """
_ClassificationOutput_ISSUE_EXAMPLE = """
            >>> output = ClassificationOutput(classes=["AF", "N", "SPB"], pred=np.ones((1,3)), prob=np.ones((1,3)), d={"d":1})
            >>> output
            {'classes': ['AF', 'N', 'SPB'],
                'prob': array([[1., 1., 1.]]),
                'pred': array([[1., 1., 1.]]),
                'd': {'d': 1}}
            >>> output.d  # has to access via `output["d"]`
            AttributeError: 'ClassificationOutput' object has no attribute 'd'
    """
_MultiLabelClassificationOutput_ISSUE_EXAMPLE = """
            >>> output = MultiLabelClassificationOutput(classes=["AF", "N", "SPB"], thr=0.5, pred=np.ones((1,3)), prob=np.ones((1,3)), d={"d":1})
            >>> output
            {'classes': ['AF', 'N', 'SPB'],
                'prob': array([[1., 1., 1.]]),
                'pred': array([[1., 1., 1.]]),
                'thr': 0.5,
                'd': {'d': 1}}
            >>> output.d  # has to access via `output["d"]`
            AttributeError: 'MultiLabelClassificationOutput' object has no attribute 'd'
    """
_SequenceTaggingOutput_ISSUE_EXAMPLE = """
            >>> output = SequenceTaggingOutput(classes=["AF", "N", "SPB"], thr=0.5, pred=np.ones((1,3,3)), prob=np.ones((1,3,3)), d={"d":1})
            >>> output
            {'classes': ['AF', 'N', 'SPB'],
                'prob': array([[[1., 1., 1.],
                        [1., 1., 1.],
                        [1., 1., 1.]]]),
                'pred': array([[[1., 1., 1.],
                        [1., 1., 1.],
                        [1., 1., 1.]]]),
                'thr': 0.5,
                'd': {'d': 1}}
            >>> output.d  # has to access via `output["d"]`
            AttributeError: 'SequenceTaggingOutput' object has no attribute 'd'
    """
_WaveDelineationOutput_ISSUE_EXAMPLE = """
            >>> output = WaveDelineationOutput(classes=["N", "P", "Q",], thr=0.5, mask=np.ones((1,3,3)), prob=np.ones((1,3,3)), d={"d":1})
            >>> output
            {'classes': ['AF', 'N', 'SPB'],
                'prob': array([[[1., 1., 1.],
                        [1., 1., 1.],
                        [1., 1., 1.]]]),
                'mask': array([[[1., 1., 1.],
                        [1., 1., 1.],
                        [1., 1., 1.]]]),
                'd': {'d': 1}}
            >>> output.d  # has to access via `output["d"]`
            AttributeError: 'WaveDelineationOutput' object has no attribute 'd'
    """
_RPeaksDetectionOutput_ISSUE_EXAMPLE = """
            >>> output = RPeaksDetectionOutput(rpeak_indices=[[2]], thr=0.5, prob=np.ones((1,3,3)), d={"d":1})
            >>> output
            {'rpeak_indices': [[2]],
                'prob': array([[[1., 1., 1.],
                        [1., 1., 1.],
                        [1., 1., 1.]]]),
                'thr': 0.5,
                'd': {'d': 1}}
            >>> output.d  # has to access via `output["d"]`
            AttributeError: 'RPeaksDetectionOutput' object has no attribute 'd'
    """


class BaseOutput(CFG, ABC):
    """Base class for all outputs.

    Parameters
    ----------
    *args : sequence
        Positional arguments.
    **kwargs : dict
        Keyword arguments.

    """

    __name__ = "BaseOutput"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        pop_fields = [
            k
            for k in self
            if k in ["required_fields", "append", "compute_metrics"]
            or k.startswith("_abc")
        ]
        for f in pop_fields:
            self.pop(f, None)
        assert all(field in self.keys() for field in self.required_fields()), (
            f"{self.__name__} requires {self.required_fields()}, "
            f"but `{', '.join(self.required_fields() - set(self.keys()))}` are missing"
        )
        assert all(
            self[field] is not None for field in self.required_fields()
        ), f"Fields `{', '.join([field for field in self.required_fields() if self[field] is None])}` are not set"

    @abstractmethod
    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        raise NotImplementedError("Subclass must implement method `required_fields`")

    def append(self, values: Union["BaseOutput", Sequence["BaseOutput"]]) -> None:
        """Append other :class:`Output` to `self`

        Parameters
        ----------
        values : Output or Sequence[Output]
            The values to be appended.

        Returns
        -------
        None

        """
        if not isinstance(values, Sequence):
            values = [values]
        for v in values:
            assert (
                v.__class__ == self.__class__
            ), "`values` must be of the same type as `self`"
            assert set(v.keys()) == set(
                self.keys()
            ), "`values` must have the same fields as `self`"
            for k, v_ in v.items():
                if k in ["classes"]:
                    assert (
                        v_ == self[k]
                    ), f"the field of ordered sequence `{k}` must be the identical"
                    continue
                if isinstance(v_, np.ndarray):
                    self[k] = np.concatenate((self[k], v_))
                elif isinstance(v_, pd.DataFrame):
                    self[k] = pd.concat([self[k], v_], axis=0, ignore_index=True)
                elif isinstance(v_, Sequence):  # list, tuple, etc.
                    self[k] += v_
                else:
                    raise ValueError(
                        f"field `{k}` of type `{type(v_)}` is not supported"
                    )


@add_docstring(_KNOWN_ISSUES.format(_ClassificationOutput_ISSUE_EXAMPLE), "append")
class ClassificationOutput(BaseOutput):
    """
    Class that maintains the output of a (typically single-label) classification task.

    Parameters
    ----------
    classes : Sequence[str]
        Class names.
    prob : numpy.ndarray
        Probabilities of each class,
        of shape ``(batch_size, num_classes)``.
    pred : numpy.ndarray
        Predicted class indices of shape ``(batch_size,)``,
        or binary predictions of shape ``(batch_size, num_classes)``.

    """

    __name__ = "ClassificationOutput"

    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        return set(
            [
                "classes",
                "prob",
                "pred",
            ]
        )

    def compute_metrics(self) -> ClassificationMetrics:
        """Compute metrics from the output.

        Returns
        -------
        metrics : ClassificationMetrics
            Metrics computed from the output.

        """
        assert hasattr(self, "labels") or hasattr(
            self, "label"
        ), "`labels` or `label` must be stored in the output for computing metrics"
        clf_met = ClassificationMetrics(multi_label=False, macro=True)
        return clf_met.compute(
            self.get("labels", self.get("label")), self.pred, len(self.classes)
        )


@add_docstring(
    _KNOWN_ISSUES.format(_MultiLabelClassificationOutput_ISSUE_EXAMPLE), "append"
)
class MultiLabelClassificationOutput(BaseOutput):
    """
    Class that maintains the output of a multi-label classification task.

    Parameters
    ----------
    classes : Sequence[str]
        class names
    thr : float
        threshold for making binary predictions
    prob : numpy.ndarray
        Probabilities of each class,
        of shape ``(batch_size, num_classes)``
    pred : numpy.ndarray
        Binary predictions,
        of shape ``(batch_size, num_classes)``.

    """

    __name__ = "MultiLabelClassificationOutput"

    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        return set(
            [
                "classes",
                "thr",
                "prob",
                "pred",
            ]
        )

    def compute_metrics(self, macro: bool = True) -> ClassificationMetrics:
        """Compute metrics from the output.

        Parameters
        ----------
        macro : bool
            Whether to use macro-averaged metrics or not.

        Returns
        -------
        metrics : ClassificationMetrics
            Metrics computed from the output.

        """
        assert hasattr(self, "labels") or hasattr(
            self, "label"
        ), "`labels` or `label` must be stored in the output for computing metrics"
        clf_met = ClassificationMetrics(multi_label=True, macro=macro)
        return clf_met.compute(
            self.get("labels", self.get("label")), self.pred, len(self.classes)
        )


@add_docstring(_KNOWN_ISSUES.format(_SequenceTaggingOutput_ISSUE_EXAMPLE), "append")
class SequenceTaggingOutput(BaseOutput):
    """Class that maintains the output of a sequence tagging task.

    Parameters
    ----------
    classes : Sequence[str]
        Class names.
    prob : numpy.ndarray
        Probabilities of each class at each time step (each sample point),
        of shape ``(batch_size, signal_length, num_classes)``.
    pred : numpy.ndarray
        Predicted class indices at each time step (each sample point),
        of shape ``(batch_size, signal_length)``;
        or binary predictions at each time step (each sample point),
        of shape ``(batch_size, signal_length, num_classes)``.

    """

    __name__ = "SequenceTaggingOutput"

    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        return set(
            [
                "classes",
                "prob",
                "pred",
            ]
        )

    def compute_metrics(self, macro: bool = True) -> ClassificationMetrics:
        """Compute metrics from the output.

        Parameters
        ----------
        macro : bool
            Whether to use macro-averaged metrics or not.

        Returns
        -------
        metrics : ClassificationMetrics
            Metrics computed from the output.

        """
        assert hasattr(self, "labels") or hasattr(
            self, "label"
        ), "`labels` or `label` must be stored in the output for computing metrics"
        clf_met = ClassificationMetrics(multi_label=False, macro=macro)
        labels = self.get("labels", self.get("label"))
        return clf_met.compute(
            labels.reshape((-1, labels.shape[-1])),
            self.pred.reshape((-1, self.pred.shape[-1])),
            len(self.classes),
        )


# alias
SequenceLabellingOutput = SequenceTaggingOutput
SequenceLabellingOutput.__name__ = "SequenceLabellingOutput"


@add_docstring(_KNOWN_ISSUES.format(_WaveDelineationOutput_ISSUE_EXAMPLE), "append")
class WaveDelineationOutput(SequenceTaggingOutput):
    """Class that maintains the output of a wave delineation task.

    Parameters
    ----------
    classes : Sequence[str]
        class names.
    prob : numpy.ndarray
        Probabilities of each class at each time step (each sample point),
        of shape ``(batch_size, signal_length, num_classes)``.
    mask : numpy.ndarray
        Predicted class indices at each time step (each sample point),
        or binary predictions at each time step (each sample point),
        of shape ``(batch_size, num_channels, signal_length)``.
    """

    __name__ = "WaveDelineationOutput"

    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        return set(
            [
                "classes",
                "prob",
                "mask",
            ]
        )

    @add_docstring(
        f"""Compute metrics from the output

        Parameters
        ----------
        fs : numbers.Real
            Sampling frequency of the signal corresponding to the masks,
            used to compute the duration of each waveform,
            and thus the error and standard deviations of errors.
        class_map : dict
            Class map, mapping names to waves to numbers from 0 to n_classes-1,
            the keys should contain {", ".join([f'"{item}"' for item in ECGWaveFormNames])}.
        macro : bool
            Whether to use macro-averaged metrics or not.
        tol : float, default 0.15
            Tolerance for the duration of the waveform,
            with units in seconds.

        Returns
        -------
        metrics : WaveDelineationMetrics
            Metrics computed from the output

        """
    )
    def compute_metrics(
        self,
        fs: int,
        class_map: Dict[str, int],
        macro: bool = True,
        tol: float = 0.15,
    ) -> ClassificationMetrics:
        assert hasattr(self, "labels") or hasattr(
            self, "label"
        ), "`labels` or `label` must be stored in the output for computing metrics"
        wd_met = WaveDelineationMetrics(macro=macro, tol=tol)
        labels = self.get("labels", self.get("label"))
        return wd_met.compute(labels, self.mask, class_map=class_map, fs=fs)


@add_docstring(_KNOWN_ISSUES.format(_RPeaksDetectionOutput_ISSUE_EXAMPLE), "append")
class RPeaksDetectionOutput(BaseOutput):
    """
    Class that maintains the output of an R peaks detection task.

    Parameters
    ----------
    rpeak_indices : Sequence[Sequence[int]]
        Rpeak indices for each batch sample.
    prob : numpy.ndarray
        Probabilities at each time step (each sample point),
        of shape ``(batch_size, signal_length)``.
    """

    __name__ = "RPeaksDetectionOutput"

    def required_fields(self) -> Set[str]:
        """The required fields of the output class."""
        return set(
            [
                "rpeak_indices",
                "prob",
            ]
        )

    def compute_metrics(self, fs: int, thr: float = 0.075) -> ClassificationMetrics:
        """Compute metrics from the output.

        Parameters
        ----------
        fs : int
            Sampling frequency of the signal corresponding to the masks.
        thr : float, default 0.075
            Threshold for a prediction to be truth positive,
            with units in seconds.

        Returns
        -------
        metrics : RPeaksDetectionMetrics
            Metrics computed from the output.

        """
        assert hasattr(self, "labels") or hasattr(
            self, "label"
        ), "`labels` or `label` must be stored in the output for computing metrics"
        rpd_met = RPeaksDetectionMetrics(thr=thr)
        return rpd_met.compute(
            self.get("labels", self.get("label")), self.rpeak_indices, fs=fs
        )
